import pytest
import numpy as np
import lenstronomy.LightModel.Profiles.ellipsoid as torus
from lenstronomy.LightModel.Profiles.ellipsoid import Ellipsoid


class TestTorus(object):
    """Class to test Shapelets."""

    def setup_method(self):
        pass

    def test_function(self):
        """

        :return:
        """
        output = torus.function(x=1, y=1, amp=1.0, sigma=2, center_x=0, center_y=0)
        assert output == 0.079577471545947673


class TestEllipsoid(object):
    def setup_method(self):
        pass

    def test_function(self):
        """

        :return:
        """
        ellipsoid = Ellipsoid()
        output = ellipsoid.function(
            x=1, y=1, amp=1.0, radius=1, e1=0, e2=0, center_x=0, center_y=0
        )
        assert output == 0
        output = ellipsoid.function(
            x=0.99, y=0, amp=1.0, radius=1, e1=0, e2=0, center_x=0, center_y=0
        )
        assert output == 1.0 / np.pi


if __name__ == "__main__":
    pytest.main()



from lenstronomy.Util.util import convert_bool_list

__all__ = ["ProfileListBase"]


_SUPPORTED_MODELS = [
    "ARC_PERT",
    "CHAMELEON",
    "CNFW",
    "CNFW_ELLIPSE",
    "CONST_MAG",
    "CONVERGENCE",
    "coreBURKERT",
    "CORED_DENSITY",
    "CORED_DENSITY_2",
    "CORED_DENSITY_2_MST",
    "CORED_DENSITY_EXP",
    "CORED_DENSITY_EXP_MST",
    "CORED_DENSITY_MST",
    "CORED_DENSITY_ULDM_MST",
    "CSE",
    "CTNFW_GAUSS_DEC",
    "CURVED_ARC_CONST",
    "CURVED_ARC_SIS_MST",
    "CURVED_ARC_SPP",
    "CURVED_ARC_SPT",
    "CURVED_ARC_TAN_DIFF",
    "DIPOLE",
    "DOUBLE_CHAMELEON",
    "EPL",
    "EPL_BOXYDISKY",
    "EPL_NUMBA",
    "EPL_Q_PHI",
    "ElliSLICE",
    "FLEXION",
    "FLEXIONFG",
    "GAUSSIAN",
    "GAUSSIAN_ELLIPSE_KAPPA",
    "GAUSSIAN_ELLIPSE_POTENTIAL",
    "GAUSSIAN_KAPPA",
    "GNFW",
    "HERNQUIST",
    "HERNQUIST_ELLIPSE",
    "HERNQUIST_ELLIPSE_CSE",
    "HESSIAN",
    "INTERPOL",
    "INTERPOL_SCALED",
    "LOS",
    "LOS_MINIMAL",
    "MULTIPOLE",
    "MULTI_GAUSSIAN_KAPPA",
    "MULTI_GAUSSIAN_KAPPA_ELLIPSE",
    "NFW",
    "NFW_ELLIPSE",
    "NFW_ELLIPSE_CSE",
    "NFW_ELLIPSE_GAUSS_DEC",
    "NFW_MC",
    "NFW_MC_ELLIPSE",
    "NIE",
    "NIE_POTENTIAL",
    "NIE_SIMPLE",
    "PEMD",
    "PJAFFE",
    "PJAFFE_ELLIPSE",
    "POINT_MASS",
    "SERSIC",
    "SERSIC_ELLIPSE_GAUSS_DEC",
    "SERSIC_ELLIPSE_KAPPA",
    "SERSIC_ELLIPSE_POTENTIAL",
    "SHAPELETS_CART",
    "SHAPELETS_POLAR",
    "SHEAR",
    "SHEAR_GAMMA_PSI",
    "SHEAR_REDUCED",
    "SHIFT",
    "SIE",
    "SIS",
    "SIS_TRUNCATED",
    "SPEMD",
    "SPEP",
    "SPL_CORE",
    "SPP",
    "SYNTHESIS",
    "TABULATED_DEFLECTIONS",
    "TNFW",
    "TNFWC",
    "TNFW_ELLIPSE",
    "TRIPLE_CHAMELEON",
    "ULDM",
]


class ProfileListBase(object):
    """Class that manages the list of lens model class instances.

    This class is applicable for single plane and multi plane lensing
    """

    def __init__(
        self,
        lens_model_list,
        numerical_alpha_class=None,
        lens_redshift_list=None,
        z_source_convention=None,
        kwargs_interp=None,
        kwargs_synthesis=None,
    ):
        """

        :param lens_model_list: list of strings with lens model names
        :param numerical_alpha_class: an instance of a custom class for use in NumericalAlpha() lens model
         deflection angles as a lens model. See the documentation in Profiles.numerical_deflections
        :param kwargs_interp: interpolation keyword arguments specifying the numerics.
         See description in the Interpolate() class. Only applicable for 'INTERPOL' and 'INTERPOL_SCALED' models.
        :param kwargs_synthesis: keyword arguments for the 'SYNTHESIS' lens model, if applicable
        """
        self.func_list = self._load_model_instances(
            lens_model_list,
            custom_class=numerical_alpha_class,
            lens_redshift_list=lens_redshift_list,
            z_source_convention=z_source_convention,
            kwargs_interp=kwargs_interp,
            kwargs_synthesis=kwargs_synthesis,
        )
        self._num_func = len(self.func_list)
        self._model_list = lens_model_list

    def _load_model_instances(
        self,
        lens_model_list,
        custom_class=None,
        lens_redshift_list=None,
        z_source_convention=None,
        kwargs_interp=None,
        kwargs_synthesis=None,
    ):
        if lens_redshift_list is None:
            lens_redshift_list = [None] * len(lens_model_list)
        if kwargs_interp is None:
            kwargs_interp = {}
        if kwargs_synthesis is None:
            kwargs_synthesis = {}
        func_list = []
        imported_classes = {}
        for i, lens_type in enumerate(lens_model_list):
            # those models require a new instance per profile as some pre-computations are different when parameters or
            # other settings are changed. For example, the 'INTERPOL' model needs to know the specific map to be
            # interpolated.
            if lens_type in [
                "NFW_MC",
                "NFW_MC_ELLIPSE",
                "CHAMELEON",
                "DOUBLE_CHAMELEON",
                "TRIPLE_CHAMELEON",
                "NFW_ELLIPSE_GAUSS_DEC",
                "CTNFW_GAUSS_DEC",
                "INTERPOL",
                "INTERPOL_SCALED",
                "NIE",
                "NIE_SIMPLE",
            ]:
                lensmodel_class = self._import_class(
                    lens_type,
                    custom_class,
                    z_lens=lens_redshift_list[i],
                    z_source=z_source_convention,
                    kwargs_interp=kwargs_interp,
                    kwargs_synthesis=kwargs_synthesis,
                )
            else:
                if lens_type not in imported_classes.keys():
                    lensmodel_class = self._import_class(
                        lens_type,
                        custom_class,
                        kwargs_interp=kwargs_interp,
                        kwargs_synthesis=kwargs_synthesis,
                    )
                    imported_classes.update({lens_type: lensmodel_class})
                else:
                    lensmodel_class = imported_classes[lens_type]
            func_list.append(lensmodel_class)
        return func_list

    @staticmethod
    def _import_class(
        lens_type,
        custom_class,
        kwargs_interp,
        kwargs_synthesis,
        z_lens=None,
        z_source=None,
    ):
        """

        :param lens_type: string, lens model type
        :param custom_class: custom class
        :param z_lens: lens redshift  # currently only used in NFW_MC model as this is redshift dependent
        :param z_source: source redshift  # currently only used in NFW_MC model as this is redshift dependent
        :param kwargs_interp: interpolation keyword arguments specifying the numerics.
         See description in the Interpolate() class. Only applicable for 'INTERPOL' and 'INTERPOL_SCALED' models.
        :return: class instance of the lens model type
        """

        if lens_type == "SHIFT":
            from lenstronomy.LensModel.Profiles.constant_shift import Shift

            return Shift()
        elif lens_type == "NIE_POTENTIAL":
            from lenstronomy.LensModel.Profiles.nie_potential import NIE_POTENTIAL

            return NIE_POTENTIAL()
        elif lens_type == "CONST_MAG":
            from lenstronomy.LensModel.Profiles.const_mag import ConstMag

            return ConstMag()
        elif lens_type == "SHEAR":
            from lenstronomy.LensModel.Profiles.shear import Shear

            return Shear()
        elif lens_type == "SHEAR_GAMMA_PSI":
            from lenstronomy.LensModel.Profiles.shear import ShearGammaPsi

            return ShearGammaPsi()
        elif lens_type == "SHEAR_REDUCED":
            from lenstronomy.LensModel.Profiles.shear import ShearReduced

            return ShearReduced()
        elif lens_type == "CONVERGENCE":
            from lenstronomy.LensModel.Profiles.convergence import Convergence

            return Convergence()
        elif lens_type == "HESSIAN":
            from lenstronomy.LensModel.Profiles.hessian import Hessian

            return Hessian()
        elif lens_type == "FLEXION":
            from lenstronomy.LensModel.Profiles.flexion import Flexion

            return Flexion()
        elif lens_type == "FLEXIONFG":
            from lenstronomy.LensModel.Profiles.flexionfg import Flexionfg

            return Flexionfg()
        elif lens_type == "POINT_MASS":
            from lenstronomy.LensModel.Profiles.point_mass import PointMass

            return PointMass()
        elif lens_type == "SIS":
            from lenstronomy.LensModel.Profiles.sis import SIS

            return SIS()
        elif lens_type == "SIS_TRUNCATED":
            from lenstronomy.LensModel.Profiles.sis_truncate import SIS_truncate

            return SIS_truncate()
        elif lens_type == "SIE":
            from lenstronomy.LensModel.Profiles.sie import SIE

            return SIE()
        elif lens_type == "SPP":
            from lenstronomy.LensModel.Profiles.spp import SPP

            return SPP()
        elif lens_type == "NIE":
            from lenstronomy.LensModel.Profiles.nie import NIE

            return NIE()
        elif lens_type == "NIE_SIMPLE":
            from lenstronomy.LensModel.Profiles.nie import NIEMajorAxis

            return NIEMajorAxis()
        elif lens_type == "CHAMELEON":
            from lenstronomy.LensModel.Profiles.chameleon import Chameleon

            return Chameleon()
        elif lens_type == "DOUBLE_CHAMELEON":
            from lenstronomy.LensModel.Profiles.chameleon import DoubleChameleon

            return DoubleChameleon()
        elif lens_type == "TRIPLE_CHAMELEON":
            from lenstronomy.LensModel.Profiles.chameleon import TripleChameleon

            return TripleChameleon()
        elif lens_type == "SPEP":
            from lenstronomy.LensModel.Profiles.spep import SPEP

            return SPEP()
        elif lens_type == "PEMD":
            from lenstronomy.LensModel.Profiles.pemd import PEMD

            return PEMD()
        elif lens_type == "SPEMD":
            from lenstronomy.LensModel.Profiles.spemd import SPEMD

            return SPEMD()
        elif lens_type == "EPL":
            from lenstronomy.LensModel.Profiles.epl import EPL

            return EPL()
        elif lens_type == "EPL_Q_PHI":
            from lenstronomy.LensModel.Profiles.epl import EPLQPhi

            return EPLQPhi()
        elif lens_type == "EPL_NUMBA":
            from lenstronomy.LensModel.Profiles.epl_numba import EPL_numba

            return EPL_numba()
        elif lens_type == "EPL_BOXYDISKY":
            from lenstronomy.LensModel.Profiles.epl_boxydisky import EPL_BOXYDISKY

            return EPL_BOXYDISKY()
        elif lens_type == "SPL_CORE":
            from lenstronomy.LensModel.Profiles.splcore import SPLCORE

            return SPLCORE()
        elif lens_type == "NFW":
            from lenstronomy.LensModel.Profiles.nfw import NFW

            return NFW()
        elif lens_type == "NFW_ELLIPSE":
            from lenstronomy.LensModel.Profiles.nfw_ellipse import NFW_ELLIPSE

            return NFW_ELLIPSE()
        elif lens_type == "NFW_ELLIPSE_GAUSS_DEC":
            from lenstronomy.LensModel.Profiles.gauss_decomposition import (
                NFWEllipseGaussDec,
            )

            return NFWEllipseGaussDec()
        elif lens_type == "NFW_ELLIPSE_CSE":
            from lenstronomy.LensModel.Profiles.nfw_ellipse_cse import NFW_ELLIPSE_CSE

            return NFW_ELLIPSE_CSE()
        elif lens_type == "TNFW":
            from lenstronomy.LensModel.Profiles.tnfw import TNFW

            return TNFW()
        elif lens_type == "TNFW_ELLIPSE":
            from lenstronomy.LensModel.Profiles.tnfw_ellipse import TNFW_ELLIPSE

            return TNFW_ELLIPSE()
        elif lens_type == "CNFW":
            from lenstronomy.LensModel.Profiles.cnfw import CNFW

            return CNFW()
        elif lens_type == "CNFW_ELLIPSE":
            from lenstronomy.LensModel.Profiles.cnfw_ellipse import CNFW_ELLIPSE

            return CNFW_ELLIPSE()
        elif lens_type == "CTNFW_GAUSS_DEC":
            from lenstronomy.LensModel.Profiles.gauss_decomposition import CTNFWGaussDec

            return CTNFWGaussDec()
        elif lens_type == "NFW_MC":
            from lenstronomy.LensModel.Profiles.nfw_mass_concentration import NFWMC

            return NFWMC(z_lens=z_lens, z_source=z_source)
        elif lens_type == "NFW_MC_ELLIPSE":
            from lenstronomy.LensModel.Profiles.nfw_mass_concentration_ellipse import (
                NFWMCEllipse,
            )

            return NFWMCEllipse(z_lens=z_lens, z_source=z_source)
        elif lens_type == "SERSIC":
            from lenstronomy.LensModel.Profiles.sersic import Sersic

            return Sersic()
        elif lens_type == "SERSIC_ELLIPSE_POTENTIAL":
            from lenstronomy.LensModel.Profiles.sersic_ellipse_potential import (
                SersicEllipse,
            )

            return SersicEllipse()
        elif lens_type == "SERSIC_ELLIPSE_KAPPA":
            from lenstronomy.LensModel.Profiles.sersic_ellipse_kappa import (
                SersicEllipseKappa,
            )

            return SersicEllipseKappa()
        elif lens_type == "SERSIC_ELLIPSE_GAUSS_DEC":
            from lenstronomy.LensModel.Profiles.gauss_decomposition import (
                SersicEllipseGaussDec,
            )

            return SersicEllipseGaussDec()
        elif lens_type == "PJAFFE":
            from lenstronomy.LensModel.Profiles.p_jaffe import PJaffe

            return PJaffe()
        elif lens_type == "PJAFFE_ELLIPSE":
            from lenstronomy.LensModel.Profiles.p_jaffe_ellipse import PJaffe_Ellipse

            return PJaffe_Ellipse()
        elif lens_type == "HERNQUIST":
            from lenstronomy.LensModel.Profiles.hernquist import Hernquist

            return Hernquist()
        elif lens_type == "HERNQUIST_ELLIPSE":
            from lenstronomy.LensModel.Profiles.hernquist_ellipse import (
                Hernquist_Ellipse,
            )

            return Hernquist_Ellipse()
        elif lens_type == "HERNQUIST_ELLIPSE_CSE":
            from lenstronomy.LensModel.Profiles.hernquist_ellipse_cse import (
                HernquistEllipseCSE,
            )

            return HernquistEllipseCSE()
        elif lens_type == "GAUSSIAN":
            from lenstronomy.LensModel.Profiles.gaussian_potential import Gaussian

            return Gaussian()
        elif lens_type == "GAUSSIAN_KAPPA":
            from lenstronomy.LensModel.Profiles.gaussian_kappa import GaussianKappa

            return GaussianKappa()
        elif lens_type == "GAUSSIAN_ELLIPSE_KAPPA":
            from lenstronomy.LensModel.Profiles.gaussian_ellipse_kappa import (
                GaussianEllipseKappa,
            )

            return GaussianEllipseKappa()
        elif lens_type == "GAUSSIAN_ELLIPSE_POTENTIAL":
            from lenstronomy.LensModel.Profiles.gaussian_ellipse_potential import (
                GaussianEllipsePotential,
            )

            return GaussianEllipsePotential()
        elif lens_type == "MULTI_GAUSSIAN_KAPPA":
            from lenstronomy.LensModel.Profiles.multi_gaussian_kappa import (
                MultiGaussianKappa,
            )

            return MultiGaussianKappa()
        elif lens_type == "MULTI_GAUSSIAN_KAPPA_ELLIPSE":
            from lenstronomy.LensModel.Profiles.multi_gaussian_kappa import (
                MultiGaussianKappaEllipse,
            )

            return MultiGaussianKappaEllipse()
        elif lens_type == "INTERPOL":
            from lenstronomy.LensModel.Profiles.interpol import Interpol

            return Interpol(**kwargs_interp)
        elif lens_type == "INTERPOL_SCALED":
            from lenstronomy.LensModel.Profiles.interpol import InterpolScaled

            return InterpolScaled(**kwargs_interp)
        elif lens_type == "SHAPELETS_POLAR":
            from lenstronomy.LensModel.Profiles.shapelet_pot_polar import PolarShapelets

            return PolarShapelets()
        elif lens_type == "SHAPELETS_CART":
            from lenstronomy.LensModel.Profiles.shapelet_pot_cartesian import (
                CartShapelets,
            )

            return CartShapelets()
        elif lens_type == "DIPOLE":
            from lenstronomy.LensModel.Profiles.dipole import Dipole

            return Dipole()
        elif lens_type == "CURVED_ARC_CONST":
            from lenstronomy.LensModel.Profiles.curved_arc_const import CurvedArcConst

            return CurvedArcConst()
        elif lens_type == "CURVED_ARC_CONST_MST":
            from lenstronomy.LensModel.Profiles.curved_arc_const import (
                CurvedArcConstMST,
            )

            return CurvedArcConstMST()
        elif lens_type == "CURVED_ARC_SPP":
            from lenstronomy.LensModel.Profiles.curved_arc_spp import CurvedArcSPP

            return CurvedArcSPP()
        elif lens_type == "CURVED_ARC_SIS_MST":
            from lenstronomy.LensModel.Profiles.curved_arc_sis_mst import (
                CurvedArcSISMST,
            )

            return CurvedArcSISMST()
        elif lens_type == "CURVED_ARC_SPT":
            from lenstronomy.LensModel.Profiles.curved_arc_spt import CurvedArcSPT

            return CurvedArcSPT()
        elif lens_type == "CURVED_ARC_TAN_DIFF":
            from lenstronomy.LensModel.Profiles.curved_arc_tan_diff import (
                CurvedArcTanDiff,
            )

            return CurvedArcTanDiff()
        elif lens_type == "ARC_PERT":
            from lenstronomy.LensModel.Profiles.arc_perturbations import (
                ArcPerturbations,
            )

            return ArcPerturbations()
        elif lens_type == "coreBURKERT":
            from lenstronomy.LensModel.Profiles.coreBurkert import CoreBurkert

            return CoreBurkert()
        elif lens_type == "CORED_DENSITY":
            from lenstronomy.LensModel.Profiles.cored_density import CoredDensity

            return CoredDensity()
        elif lens_type == "CORED_DENSITY_2":
            from lenstronomy.LensModel.Profiles.cored_density_2 import CoredDensity2

            return CoredDensity2()
        elif lens_type == "CORED_DENSITY_EXP":
            from lenstronomy.LensModel.Profiles.cored_density_exp import CoredDensityExp

            return CoredDensityExp()
        elif lens_type == "CORED_DENSITY_MST":
            from lenstronomy.LensModel.Profiles.cored_density_mst import CoredDensityMST

            return CoredDensityMST(profile_type="CORED_DENSITY")
        elif lens_type == "CORED_DENSITY_2_MST":
            from lenstronomy.LensModel.Profiles.cored_density_mst import CoredDensityMST

            return CoredDensityMST(profile_type="CORED_DENSITY_2")
        elif lens_type == "CORED_DENSITY_EXP_MST":
            from lenstronomy.LensModel.Profiles.cored_density_mst import CoredDensityMST

            return CoredDensityMST(profile_type="CORED_DENSITY_EXP")
        elif lens_type == "TABULATED_DEFLECTIONS":
            from lenstronomy.LensModel.Profiles.numerical_deflections import (
                TabulatedDeflections,
            )

            return TabulatedDeflections(custom_class)
        elif lens_type == "MULTIPOLE":
            from lenstronomy.LensModel.Profiles.multipole import Multipole

            return Multipole()
        elif lens_type == "CSE":
            from lenstronomy.LensModel.Profiles.cored_steep_ellipsoid import CSE

            return CSE()
        elif lens_type == "ElliSLICE":
            from lenstronomy.LensModel.Profiles.elliptical_density_slice import (
                ElliSLICE,
            )

            return ElliSLICE()
        elif lens_type == "ULDM":
            from lenstronomy.LensModel.Profiles.uldm import Uldm

            return Uldm()
        elif lens_type == "GNFW":
            from lenstronomy.LensModel.Profiles.general_nfw import GNFW

            return GNFW()
        elif lens_type == "CORED_DENSITY_ULDM_MST":
            from lenstronomy.LensModel.Profiles.cored_density_mst import CoredDensityMST

            return CoredDensityMST(profile_type="CORED_DENSITY_ULDM")
        elif lens_type == "LOS":
            from lenstronomy.LensModel.LineOfSight.LOSModels.los import LOS

            return LOS()
        elif lens_type == "LOS_MINIMAL":
            from lenstronomy.LensModel.LineOfSight.LOSModels.los_minimal import (
                LOSMinimal,
            )

            return LOSMinimal()
        elif lens_type == "SYNTHESIS":
            from lenstronomy.LensModel.Profiles.synthesis import SynthesisProfile

            return SynthesisProfile(**kwargs_synthesis)
        elif lens_type == "TNFWC":
            from lenstronomy.LensModel.Profiles.nfw_core_truncated import TNFWC

            return TNFWC()
        else:
            raise ValueError(
                "%s is not a valid lens model. Supported are: %s."
                % (lens_type, _SUPPORTED_MODELS)
            )

    def _bool_list(self, k=None):
        """Returns a bool list of the length of the lens models if k = None: returns
        bool list with True's if k is int, returns bool list with False's but k'th is
        True if k is a list of int, e.g. [0, 3, 5], returns a bool list with True's in
        the integers listed and False elsewhere if k is a boolean list, checks for size
        to match the numbers of models and returns it.

        :param k: None, int, or list of ints
        :return: bool list
        """
        return convert_bool_list(n=self._num_func, k=k)

    def set_static(self, kwargs_list):
        """

        :param kwargs_list: list of keyword arguments for each profile
        :return: kwargs_list
        """
        for i, func in enumerate(self.func_list):
            func.set_static(**kwargs_list[i])
        return kwargs_list

    def set_dynamic(self):
        """Frees cache set by static model (if exists) and re-computes all lensing
        quantities each time a definition is called assuming different parameters are
        executed. This is the default mode if not specified as set_static()

        :return: None
        """
        for i, func in enumerate(self.func_list):
            func.set_dynamic()



from lenstronomy.ImSim.image_model import ImageModel
import lenstronomy.ImSim.de_lens as de_lens
from lenstronomy.Util import util
from lenstronomy.ImSim.Numerics.convolution import PixelKernelConvolution
import numpy as np

__all__ = ["ImageLinearFit"]


class ImageLinearFit(ImageModel):
    """Linear version class, inherits ImageModel.

    When light models use pixel-based profile types, such as 'SLIT_STARLETS', the WLS
    linear inversion is replaced by the regularized inversion performed by an external
    solver. The current pixel-based solver is provided by the SLITronomy plug-in.
    """

    def __init__(
        self,
        data_class,
        psf_class=None,
        lens_model_class=None,
        source_model_class=None,
        lens_light_model_class=None,
        point_source_class=None,
        extinction_class=None,
        kwargs_numerics=None,
        likelihood_mask=None,
        psf_error_map_bool_list=None,
        kwargs_pixelbased=None,
        linear_solver=True,
    ):
        """

        :param data_class: ImageData() instance
        :param psf_class: PSF() instance
        :param lens_model_class: LensModel() instance
        :param source_model_class: LightModel() instance
        :param lens_light_model_class: LightModel() instance
        :param point_source_class: PointSource() instance
        :param kwargs_numerics: keyword arguments passed to the Numerics module
        :param likelihood_mask: 2d boolean array of pixels to be counted in the likelihood calculation/linear
         optimization
        :param psf_error_map_bool_list: list of boolean of length of point source models.
         Indicates whether PSF error map is used for the point source model stated as the index.
        :param kwargs_pixelbased: keyword arguments with various settings related to the pixel-based solver
         (see SLITronomy documentation) being applied to the point sources.
        :param linear_solver: bool, if True (default) fixes the linear amplitude parameters 'amp' (avoid sampling) such
         that they get overwritten by the linear solver solution.
        """
        super(ImageLinearFit, self).__init__(
            data_class,
            psf_class=psf_class,
            lens_model_class=lens_model_class,
            source_model_class=source_model_class,
            lens_light_model_class=lens_light_model_class,
            point_source_class=point_source_class,
            extinction_class=extinction_class,
            kwargs_numerics=kwargs_numerics,
            kwargs_pixelbased=kwargs_pixelbased,
        )
        self._linear_solver = linear_solver
        if psf_error_map_bool_list is None:
            psf_error_map_bool_list = [True] * len(
                self.PointSource.point_source_type_list
            )
        self._psf_error_map_bool_list = psf_error_map_bool_list
        if likelihood_mask is None:
            likelihood_mask = np.ones_like(data_class.data)
        self.likelihood_mask = np.array(likelihood_mask, dtype=bool)
        self._mask1d = util.image2array(self.likelihood_mask)
        if self._pixelbased_bool is True:
            # update the pixel-based solver with the likelihood mask
            self.PixelSolver.set_likelihood_mask(self.likelihood_mask)

        # prepare to use fft convolution for the natwt linear solver
        if self.Data.likelihood_method() == "interferometry_natwt":
            self._convolution = PixelKernelConvolution(
                kernel=self.PSF.kernel_point_source
            )

    def image_linear_solve(
        self,
        kwargs_lens=None,
        kwargs_source=None,
        kwargs_lens_light=None,
        kwargs_ps=None,
        kwargs_extinction=None,
        kwargs_special=None,
        inv_bool=False,
    ):
        """Computes the image (lens and source surface brightness with a given lens
        model). The linear parameters are computed with a weighted linear least square
        optimization (i.e. flux normalization of the brightness profiles) However in
        case of pixel-based modelling, pixel values are constrained by an external
        solver (e.g. SLITronomy).

        :param kwargs_lens: list of keyword arguments corresponding to the superposition
            of different lens profiles
        :param kwargs_source: list of keyword arguments corresponding to the
            superposition of different source light profiles
        :param kwargs_lens_light: list of keyword arguments corresponding to different
            lens light surface brightness profiles
        :param kwargs_ps: keyword arguments corresponding to "other" parameters, such as
            external shear and point source image positions
        :param inv_bool: if True, invert the full linear solver Matrix Ax = y for the
            purpose of the covariance matrix.
        :return: 2d array of surface brightness pixels of the optimal solution of the
            linear parameters to match the data
        """
        return self._image_linear_solve(
            kwargs_lens,
            kwargs_source,
            kwargs_lens_light,
            kwargs_ps,
            kwargs_extinction,
            kwargs_special,
            inv_bool=inv_bool,
        )

    def _image_linear_solve(
        self,
        kwargs_lens=None,
        kwargs_source=None,
        kwargs_lens_light=None,
        kwargs_ps=None,
        kwargs_extinction=None,
        kwargs_special=None,
        inv_bool=False,
    ):
        """Computes the image (lens and source surface brightness with a given lens
        model). By default, the linear parameters are computed with a weighted linear
        least square optimization (i.e. flux normalization of the brightness profiles)
        However in case of pixel-based modelling, pixel values are constrained by an
        external solver (e.g. SLITronomy).

        :param kwargs_lens: list of keyword arguments corresponding to the superposition
            of different lens profiles
        :param kwargs_source: list of keyword arguments corresponding to the
            superposition of different source light profiles
        :param kwargs_lens_light: list of keyword arguments corresponding to different
            lens light surface brightness profiles
        :param kwargs_ps: keyword arguments corresponding to "other" parameters, such as
            external shear and point source image positions
        :param inv_bool: if True, invert the full linear solver Matrix Ax = y for the
            purpose of the covariance matrix. This has no impact in case of pixel-based
            modelling.
        :return: 2d array of surface brightness pixels of the optimal solution of the
            linear parameters to match the data
        """
        if self._pixelbased_bool is True:
            model, model_error, cov_param, param = self.image_pixelbased_solve(
                kwargs_lens,
                kwargs_source,
                kwargs_lens_light,
                kwargs_ps,
                kwargs_extinction,
                kwargs_special,
            )
        elif self.Data.likelihood_method() == "diagonal":
            A = self._linear_response_matrix(
                kwargs_lens,
                kwargs_source,
                kwargs_lens_light,
                kwargs_ps,
                kwargs_extinction,
                kwargs_special,
            )
            C_D_response, model_error = self._error_response(
                kwargs_lens, kwargs_ps, kwargs_special=kwargs_special
            )
            d = self.data_response
            param, cov_param, wls_model = de_lens.get_param_WLS(
                A.T, 1 / C_D_response, d, inv_bool=inv_bool
            )
            model = self.array_masked2image(wls_model)
            _, _, _, _ = self._update_linear_kwargs(
                param, kwargs_lens, kwargs_source, kwargs_lens_light, kwargs_ps
            )
        elif self.Data.likelihood_method() == "interferometry_natwt":
            (
                model,
                model_error,
                cov_param,
                param,
            ) = self._image_linear_solve_interferometry_natwt(
                kwargs_lens,
                kwargs_source,
                kwargs_lens_light,
                kwargs_ps,
                kwargs_extinction,
                kwargs_special,
            )
        else:
            raise ValueError(
                "likelihood_method %s not supported!" % self.Data.likelihood_method()
            )
        return model, model_error, cov_param, param

    def image_pixelbased_solve(
        self,
        kwargs_lens=None,
        kwargs_source=None,
        kwargs_lens_light=None,
        kwargs_ps=None,
        kwargs_extinction=None,
        kwargs_special=None,
        init_lens_light_model=None,
    ):
        """Computes the image (lens and source surface brightness with a given lens
        model) using the pixel-based solver.

        :param kwargs_lens: list of keyword arguments corresponding to the superposition
            of different lens profiles
        :param kwargs_source: list of keyword arguments corresponding to the
            superposition of different source light profiles
        :param kwargs_lens_light: list of keyword arguments corresponding to different
            lens light surface brightness profiles
        :param kwargs_ps: keyword arguments corresponding to point sources
        :param kwargs_extinction: keyword arguments corresponding to dust extinction
        :param kwargs_special: keyword arguments corresponding to "special" parameters
        :param init_lens_light_model: optional initial guess for the lens surface
            brightness
        :return: 2d array of surface brightness pixels of the optimal solution of the
            linear parameters to match the data
        """
        _, model_error = self._error_response(
            kwargs_lens, kwargs_ps, kwargs_special=kwargs_special
        )
        model, param, _ = self.PixelSolver.solve(
            kwargs_lens,
            kwargs_source,
            kwargs_lens_light=kwargs_lens_light,
            kwargs_ps=kwargs_ps,
            kwargs_special=kwargs_special,
            init_lens_light_model=init_lens_light_model,
        )
        cov_param = None
        _, _ = self.update_pixel_kwargs(kwargs_source, kwargs_lens_light)
        _, _, _, _ = self._update_linear_kwargs(
            param, kwargs_lens, kwargs_source, kwargs_lens_light, kwargs_ps
        )
        return model, model_error, cov_param, param

    def linear_response_matrix(
        self,
        kwargs_lens=None,
        kwargs_source=None,
        kwargs_lens_light=None,
        kwargs_ps=None,
        kwargs_extinction=None,
        kwargs_special=None,
    ):
        """Computes the linear response matrix (m x n), with n being the data size and m
        being the coefficients.

        :param kwargs_lens: lens model keyword argument list
        :param kwargs_source: extended source model keyword argument list
        :param kwargs_lens_light: lens light model keyword argument list
        :param kwargs_ps: point source model keyword argument list
        :param kwargs_extinction: extinction model keyword argument list
        :param kwargs_special: special keyword argument list
        :return: linear response matrix
        """
        A = self._linear_response_matrix(
            kwargs_lens,
            kwargs_source,
            kwargs_lens_light,
            kwargs_ps,
            kwargs_extinction,
            kwargs_special,
        )
        return A

    @property
    def data_response(self):
        """Returns the 1d array of the data element that is fitted for (including
        masking)

        :return: 1d numpy array
        """
        d = self.image2array_masked(self.Data.data)
        return d

    def error_response(self, kwargs_lens, kwargs_ps, kwargs_special):
        """Returns the 1d array of the error estimate corresponding to the data
        response.

        :return: 1d numpy array of response, 2d array of additional errors (e.g. point
            source uncertainties)
        """
        return self._error_response(
            kwargs_lens, kwargs_ps, kwargs_special=kwargs_special
        )

    def _error_response(self, kwargs_lens, kwargs_ps, kwargs_special):
        """Returns the 1d array of the error estimate corresponding to the data
        response.

        :return: 1d numpy array of response, 2d array of additional errors (e.g. point
            source uncertainties)
        """
        model_error = self._error_map_model(
            kwargs_lens, kwargs_ps, kwargs_special=kwargs_special
        )
        # adding the uncertainties estimated from the data with the ones from the model
        C_D_response = self.image2array_masked(self.Data.C_D + model_error)
        return C_D_response, model_error

    def likelihood_data_given_model(
        self,
        kwargs_lens=None,
        kwargs_source=None,
        kwargs_lens_light=None,
        kwargs_ps=None,
        kwargs_extinction=None,
        kwargs_special=None,
        source_marg=False,
        linear_prior=None,
        check_positive_flux=False,
        linear_solver=True,
    ):
        """Computes the likelihood of the data given a model This is specified with the
        non-linear parameters and a linear inversion and prior marginalisation.

        :param kwargs_lens: list of keyword arguments corresponding to the superposition
            of different lens profiles
        :param kwargs_source: list of keyword arguments corresponding to the
            superposition of different source light profiles
        :param kwargs_lens_light: list of keyword arguments corresponding to different
            lens light surface brightness profiles
        :param kwargs_ps: keyword arguments corresponding to "other" parameters, such as
            external shear and point source image positions
        :param kwargs_extinction:
        :param kwargs_special:
        :param source_marg: bool, performs a marginalization over the linear parameters
        :param linear_prior: linear prior width in eigenvalues
        :param check_positive_flux: bool, if True, checks whether the linear inversion
            resulted in non-negative flux components and applies a punishment in the
            likelihood if so.
        :param linear_solver: bool, if True (default) fixes the linear amplitude
            parameters 'amp' (avoid sampling) such that they get overwritten by the
            linear solver solution.
        :return: log likelihood (natural logarithm)
        """
        return self._likelihood_data_given_model(
            kwargs_lens,
            kwargs_source,
            kwargs_lens_light,
            kwargs_ps,
            kwargs_extinction,
            kwargs_special,
            source_marg,
            linear_prior=linear_prior,
            check_positive_flux=check_positive_flux,
            linear_solver=linear_solver,
        )

    def _likelihood_data_given_model(
        self,
        kwargs_lens=None,
        kwargs_source=None,
        kwargs_lens_light=None,
        kwargs_ps=None,
        kwargs_extinction=None,
        kwargs_special=None,
        source_marg=False,
        linear_prior=None,
        check_positive_flux=False,
        linear_solver=True,
    ):
        """Computes the likelihood of the data given a model This is specified with the
        non-linear parameters and a linear inversion and prior marginalisation.

        :param kwargs_lens: list of keyword arguments corresponding to the superposition
            of different lens profiles
        :param kwargs_source: list of keyword arguments corresponding to the
            superposition of different source light profiles
        :param kwargs_lens_light: list of keyword arguments corresponding to different
            lens light surface brightness profiles
        :param kwargs_ps: keyword arguments corresponding to "other" parameters, such as
            external shear and point source image positions
        :param source_marg: bool, performs a marginalization over the linear parameters
        :param linear_prior: linear prior width in eigenvalues
        :param check_positive_flux: bool, if True, checks whether the linear inversion
            resulted in non-negative flux components and applies a punishment in the
            likelihood if so.
        :param linear_solver: bool, if True (default) fixes the linear amplitude
            parameters 'amp' (avoid sampling) such that they get overwritten by the
            linear solver solution.
        :return: log likelihood (natural logarithm)
        """
        # generate image
        if linear_solver is False:
            im_sim = self.image(
                kwargs_lens,
                kwargs_source,
                kwargs_lens_light,
                kwargs_ps,
                kwargs_extinction,
                kwargs_special,
            )
            cov_matrix = None
            model_error = self._error_map_model(
                kwargs_lens, kwargs_ps=kwargs_ps, kwargs_special=kwargs_special
            )
        else:
            im_sim, model_error, cov_matrix, param = self._image_linear_solve(
                kwargs_lens,
                kwargs_source,
                kwargs_lens_light,
                kwargs_ps,
                kwargs_extinction,
                kwargs_special,
                inv_bool=source_marg,
            )
        # compute X^2
        logL = self.Data.log_likelihood(im_sim, self.likelihood_mask, model_error)

        if self._pixelbased_bool is False:
            if cov_matrix is not None and source_marg:
                marg_const = de_lens.marginalization_new(
                    cov_matrix, d_prior=linear_prior
                )
                logL += marg_const
        if check_positive_flux is True:
            bool_ = self.check_positive_flux(
                kwargs_source, kwargs_lens_light, kwargs_ps
            )
            if bool_ is False:
                logL -= 10**8
        return logL

    def num_param_linear(
        self, kwargs_lens, kwargs_source, kwargs_lens_light, kwargs_ps
    ):
        """

        :return: number of linear coefficients to be solved for in the linear inversion
        """
        return self._num_param_linear(
            kwargs_lens, kwargs_source, kwargs_lens_light, kwargs_ps
        )

    def _num_param_linear(
        self, kwargs_lens, kwargs_source, kwargs_lens_light, kwargs_ps
    ):
        """

        :return: number of linear coefficients to be solved for in the linear inversion
        """
        num = 0
        if self._pixelbased_bool is False:
            num += self.SourceModel.num_param_linear(kwargs_source)
            num += self.LensLightModel.num_param_linear(kwargs_lens_light)
        num += self.PointSource.num_basis(kwargs_ps, kwargs_lens)
        return num

    def _linear_response_matrix(
        self,
        kwargs_lens,
        kwargs_source,
        kwargs_lens_light,
        kwargs_ps,
        kwargs_extinction=None,
        kwargs_special=None,
        unconvolved=False,
    ):
        """Computes the linear response matrix (m x n), with n being the data size and m
        being the coefficients.

        The calculation is done by
        - first (optional) computing differential extinctions
        - adding linear components of the lensed source(s)
        - adding linear components of the unlensed components (i.e. deflector)
        - adding point sources (can be multiple lensed or stars in the field)

        :param kwargs_lens: list of keyword arguments corresponding to the superposition of different lens profiles
        :param kwargs_source: list of keyword arguments corresponding to the superposition of different source light profiles
        :param kwargs_lens_light: list of keyword arguments corresponding to different lens light surface brightness profiles
        :param kwargs_ps: keyword arguments corresponding to "other" parameters, such as external shear and point source image positions
        :param unconvolved: bool, if True, computes components without convolution kernel (will not work for point sources)
        :return: response matrix (m x n)
        """
        x_grid, y_grid = self.ImageNumerics.coordinates_evaluate
        source_light_response, n_source = self.source_mapping.image_flux_split(
            x_grid, y_grid, kwargs_lens, kwargs_source
        )
        extinction = self._extinction.extinction(
            x_grid,
            y_grid,
            kwargs_extinction=kwargs_extinction,
            kwargs_special=kwargs_special,
        )
        lens_light_response, n_lens_light = self.LensLightModel.functions_split(
            x_grid, y_grid, kwargs_lens_light
        )

        ra_pos, dec_pos, amp, n_points = self.point_source_linear_response_set(
            kwargs_ps, kwargs_lens, kwargs_special, with_amp=False
        )
        num_param = n_points + n_lens_light + n_source

        num_response = self.num_data_evaluate
        A = np.zeros((num_param, num_response))
        n = 0
        # response of lensed source profile
        for i in range(0, n_source):
            image = source_light_response[i]

            # multiply with primary beam before convolution
            if self._pb is not None:
                image *= self._pb_1d

            image *= extinction
            image = self.ImageNumerics.re_size_convolve(image, unconvolved=unconvolved)
            A[n, :] = np.nan_to_num(self.image2array_masked(image), copy=False)
            n += 1
        # response of deflector light profile (or any other un-lensed extended components)
        for i in range(0, n_lens_light):
            image = lens_light_response[i]

            # multiply with primary beam before convolution
            if self._pb is not None:
                image *= self._pb_1d

            image = self.ImageNumerics.re_size_convolve(image, unconvolved=unconvolved)
            A[n, :] = np.nan_to_num(self.image2array_masked(image), copy=False)
            n += 1
        # response of point sources
        for i in range(0, n_points):
            # raise warnings when primary beam is attempted to be applied for point sources
            if self._pb is not None:
                raise Warning("Antenna primary beam does not apply to point sources!")

            image = self.ImageNumerics.point_source_rendering(
                ra_pos[i], dec_pos[i], amp[i]
            )
            A[n, :] = np.nan_to_num(self.image2array_masked(image), copy=False)
            n += 1
        return A * self._flux_scaling

    def update_linear_kwargs(
        self, param, kwargs_lens, kwargs_source, kwargs_lens_light, kwargs_ps
    ):
        """Links linear parameters to kwargs arguments.

        :param param: linear parameter vector corresponding to the response matrix
        :return: updated list of kwargs with linear parameter values
        """
        return self._update_linear_kwargs(
            param, kwargs_lens, kwargs_source, kwargs_lens_light, kwargs_ps
        )

    def _update_linear_kwargs(
        self, param, kwargs_lens, kwargs_source, kwargs_lens_light, kwargs_ps
    ):
        """Links linear parameters to kwargs arguments.

        :param param: linear parameter vector corresponding to the response matrix
        :return: updated list of kwargs with linear parameter values
        """
        i = 0
        kwargs_source, i = self.SourceModel.update_linear(
            param, i, kwargs_list=kwargs_source
        )
        kwargs_lens_light, i = self.LensLightModel.update_linear(
            param, i, kwargs_list=kwargs_lens_light
        )
        kwargs_ps, i = self.PointSource.update_linear(param, i, kwargs_ps, kwargs_lens)
        return kwargs_lens, kwargs_source, kwargs_lens_light, kwargs_ps

    def linear_param_from_kwargs(self, kwargs_source, kwargs_lens_light, kwargs_ps):
        """Inverse function of update_linear() returning the linear amplitude list for
        the keyword argument list.

        :param kwargs_source:
        :param kwargs_lens_light:
        :param kwargs_ps:
        :return: list of linear coefficients
        """
        return self._linear_param_from_kwargs(
            kwargs_source, kwargs_lens_light, kwargs_ps
        )

    def _linear_param_from_kwargs(self, kwargs_source, kwargs_lens_light, kwargs_ps):
        """Inverse function of update_linear() returning the linear amplitude list for
        the keyword argument list.

        :param kwargs_source:
        :param kwargs_lens_light:
        :param kwargs_ps:
        :return: list of linear coefficients
        """
        param = []
        param += self.SourceModel.linear_param_from_kwargs(kwargs_source)
        param += self.LensLightModel.linear_param_from_kwargs(kwargs_lens_light)
        param += self.PointSource.linear_param_from_kwargs(kwargs_ps)
        return param

    def update_pixel_kwargs(self, kwargs_source, kwargs_lens_light):
        """Update kwargs arguments for pixel-based profiles with fixed properties such
        as their number of pixels, scale, and center coordinates (fixed to the origin).

        :param kwargs_source: list of keyword arguments corresponding to the
            superposition of different source light profiles
        :param kwargs_lens_light: list of keyword arguments corresponding to the
            superposition of different lens light profiles
        :return: updated kwargs_source and kwargs_lens_light
        """
        # in case the source plane grid size has changed, update the kwargs accordingly
        ss_factor_source = self.SourceNumerics.grid_supersampling_factor
        kwargs_source[0]["n_pixels"] = int(
            self.Data.num_pixel * ss_factor_source**2
        )  #  effective number of pixels in source plane
        kwargs_source[0]["scale"] = (
            self.Data.pixel_width / ss_factor_source
        )  # effective pixel size of source plane grid
        # pixelated reconstructions have no well-defined center, we put it arbitrarily at (0, 0), center of the image
        kwargs_source[0]["center_x"] = 0
        kwargs_source[0]["center_y"] = 0
        # do the same if the lens light has been reconstructed
        if kwargs_lens_light is not None and len(kwargs_lens_light) > 0:
            kwargs_lens_light[0]["n_pixels"] = self.Data.num_pixel
            kwargs_lens_light[0]["scale"] = self.Data.pixel_width
            kwargs_lens_light[0]["center_x"] = 0
            kwargs_lens_light[0]["center_y"] = 0
        return kwargs_source, kwargs_lens_light

    def reduced_residuals(self, model, error_map=0):
        """

        :param model: 2d numpy array of the modeled image
        :param error_map: 2d numpy array of additional noise/error terms from model components (such as PSF model uncertainties)
        :return: 2d numpy array of reduced residuals per pixel
        """
        mask = self.likelihood_mask
        C_D = self.Data.C_D_model(model)
        residual = (model - self.Data.data) / np.sqrt(C_D + np.abs(error_map)) * mask
        return residual

    def reduced_chi2(self, model, error_map=0):
        """Returns reduced chi2 :param model: 2d numpy array of a model predicted image
        :param error_map: same format as model, additional error component (such as PSF
        errors) :return: reduced chi2."""
        norm_res = self.reduced_residuals(model, error_map)
        return np.sum(norm_res**2) / self.num_data_evaluate

    @property
    def num_data_evaluate(self):
        """Number of data points to be used in the linear solver :return: number of
        evaluated data points :rtype: int."""
        return int(np.sum(self.likelihood_mask))

    def update_data(self, data_class):
        """

        :param data_class: instance of Data() class
        :return: no return. Class is updated.
        """
        self.Data = data_class
        self.ImageNumerics._PixelGrid = data_class

    def image2array_masked(self, image):
        """Returns 1d array of values in image that are not masked out for the
        likelihood computation/linear minimization :param image: 2d numpy array of full
        image :return: 1d array."""
        array = util.image2array(image)
        return array[self._mask1d]

    def array_masked2image(self, array):
        """

        :param array: 1d array of values not masked out (part of linear fitting)
        :return: 2d array of full image
        """
        nx, ny = self.Data.num_pixel_axes
        grid1d = np.zeros(nx * ny)
        grid1d[self._mask1d] = array
        grid2d = util.array2image(grid1d, nx, ny)
        return grid2d

    def _error_map_model(self, kwargs_lens, kwargs_ps, kwargs_special=None):
        """Noise estimate (variances as diagonal of the pixel covariance matrix)
        resulted from inherent model uncertainties This term is currently the psf error
        map.

        :param kwargs_lens: lens model keyword arguments
        :param kwargs_ps: point source keyword arguments
        :param kwargs_special: special parameter keyword arguments
        :return: 2d array corresponding to the pixels in terms of variance in noise
        """
        return self._error_map_psf(kwargs_lens, kwargs_ps, kwargs_special)

    def _error_map_psf(self, kwargs_lens, kwargs_ps, kwargs_special=None):
        """Map of image with error terms (sigma**2) expected from inaccuracies in the
        PSF modeling.

        :param kwargs_lens: lens model keyword arguments
        :param kwargs_ps: point source keyword arguments
        :param kwargs_special: special parameter keyword arguments
        :return: 2d array of size of the image
        """
        error_map = np.zeros(self.Data.num_pixel_axes)
        if self._psf_error_map is True:
            for k, bool_ in enumerate(self._psf_error_map_bool_list):
                if bool_ is True:
                    ra_pos, dec_pos, _ = self.PointSource.point_source_list(
                        kwargs_ps, kwargs_lens=kwargs_lens, k=k, with_amp=False
                    )
                    if len(ra_pos) > 0:
                        ra_pos, dec_pos = self._displace_astrometry(
                            ra_pos, dec_pos, kwargs_special=kwargs_special
                        )
                        error_map += self.ImageNumerics.psf_error_map(
                            ra_pos,
                            dec_pos,
                            None,
                            self.Data.data,
                            fix_psf_error_map=False,
                        )
        return error_map

    def error_map_source(self, kwargs_source, x_grid, y_grid, cov_param):
        """Variance of the linear source reconstruction in the source plane coordinates,
        computed by the diagonal elements of the covariance matrix of the source
        reconstruction as a sum of the errors of the basis set.

        :param kwargs_source: keyword arguments of source model
        :param x_grid: x-axis of positions to compute error map
        :param y_grid: y-axis of positions to compute error map
        :param cov_param: covariance matrix of liner inversion parameters
        :return: diagonal covariance errors at the positions (x_grid, y_grid)
        """
        return self._error_map_source(kwargs_source, x_grid, y_grid, cov_param)

    def _error_map_source(self, kwargs_source, x_grid, y_grid, cov_param):
        """Variance of the linear source reconstruction in the source plane coordinates,
        computed by the diagonal elements of the covariance matrix of the source
        reconstruction as a sum of the errors of the basis set.

        :param kwargs_source: keyword arguments of source model
        :param x_grid: x-axis of positions to compute error map
        :param y_grid: y-axis of positions to compute error map
        :param cov_param: covariance matrix of liner inversion parameters
        :return: diagonal covariance errors at the positions (x_grid, y_grid)
        """

        error_map = np.zeros_like(x_grid)
        basis_functions, n_source = self.SourceModel.functions_split(
            x_grid, y_grid, kwargs_source
        )
        basis_functions = np.array(basis_functions)

        if cov_param is not None:
            for i in range(len(error_map)):
                error_map[i] = (
                    basis_functions[:, i]
                    .T.dot(cov_param[:n_source, :n_source])
                    .dot(basis_functions[:, i])
                )
        return error_map

    def point_source_linear_response_set(
        self, kwargs_ps, kwargs_lens, kwargs_special, with_amp=True
    ):
        """

        :param kwargs_ps: point source keyword argument list
        :param kwargs_lens: lens model keyword argument list
        :param kwargs_special: special keyword argument list, may include 'delta_x_image' and 'delta_y_image'
        :param with_amp: bool, if True, relative magnification between multiply imaged point sources are held fixed.
        :return: list of positions and amplitudes split in different basis components with applied astrometric corrections
        """

        ra_pos, dec_pos, amp, n_points = self.PointSource.linear_response_set(
            kwargs_ps, kwargs_lens, with_amp=with_amp
        )

        if kwargs_special is not None:
            if "delta_x_image" in kwargs_special:
                delta_x, delta_y = (
                    kwargs_special["delta_x_image"],
                    kwargs_special["delta_y_image"],
                )
                k = 0
                n = len(delta_x)
                for i in range(n_points):
                    for j in range(len(ra_pos[i])):
                        if k >= n:
                            break
                        ra_pos[i][j] = ra_pos[i][j] + delta_x[k]
                        dec_pos[i][j] = dec_pos[i][j] + delta_y[k]
                        k += 1
        return ra_pos, dec_pos, amp, n_points

    def check_positive_flux(self, kwargs_source, kwargs_lens_light, kwargs_ps):
        """Checks whether the surface brightness profiles contain positive fluxes and
        returns bool if True.

        :param kwargs_source: source surface brightness keyword argument list
        :param kwargs_lens_light: lens surface brightness keyword argument list
        :param kwargs_ps: point source keyword argument list
        :return: boolean
        """
        pos_bool_ps = self.PointSource.check_positive_flux(kwargs_ps)
        if self._pixelbased_bool is True:
            # this constraint must be handled by the pixel-based solver
            pos_bool_source = True
            pos_bool_lens_light = True
        else:
            pos_bool_source = self.SourceModel.check_positive_flux_profile(
                kwargs_source
            )
            pos_bool_lens_light = self.LensLightModel.check_positive_flux_profile(
                kwargs_lens_light
            )
        if (
            pos_bool_ps is True
            and pos_bool_source is True
            and pos_bool_lens_light is True
        ):
            return True
        else:
            return False

    # linear solver for interferometric natwt method
    def _image_linear_solve_interferometry_natwt(
        self,
        kwargs_lens=None,
        kwargs_source=None,
        kwargs_lens_light=None,
        kwargs_ps=None,
        kwargs_extinction=None,
        kwargs_special=None,
    ):
        """'interferometry_natwt' method does NOT support model_error, cov_param. The
        interferometry linear solver just does the linear solving to get the optimal
        linear amplitudes and apply the marginalized amplitudes to make the model
        images.

        :return: model, model_error, cov_param, param
        model and param are the same returns of self._image_linear_solve_interferometry_natwt_solving(A, d) function
        model_error =0 and cov_param = None for the interferometric method.
        """
        A = self._linear_response_matrix(
            kwargs_lens,
            kwargs_source,
            kwargs_lens_light,
            kwargs_ps,
            kwargs_extinction,
            kwargs_special,
            unconvolved=True,
        )
        d = self.data_response
        model, param = self._image_linear_solve_interferometry_natwt_solving(A, d)
        model_error = 0  # just a place holder
        cov_param = None  # just a place holder
        _, _, _, _ = self._update_linear_kwargs(
            param, kwargs_lens, kwargs_source, kwargs_lens_light, kwargs_ps
        )
        return model, model_error, cov_param, param

    def _image_linear_solve_interferometry_natwt_solving(self, A, d):
        """Linearly solve the amplitude of each light profile response to the natural
        weighting interferometry images, based on (placeholder for Nan Zhang's paper).

        Theories:
            Suppose there are a set of light responses :math:`\\{x_i\\}`, we want to solve the set of amplitudes :math:`\\{\\alpha_i\\}`,
            such that minimizes the chi^2 given by
            .. math::
                    \\chi^2 = (d - A_{PSF}\\sum_i \\alpha_i x_i)^TC^{-1}(d - A_{PSF}\\sum_i \\alpha_i x_i),
            where :math:`A_{PSF}` is the PSF convolution operation matrix (not to be confused with the input A of this function)
            and :math:`C` is the noise covariance matrix. :math:`d` is the data image.
            For natural weighting interferometric images, we have :math:`C = \\sigma^2 A_{PSF}`,
            (see Section 3.2 of https://doi.org/10.1093/mnras/staa2740 for the relation of natural weighting covariance matrix and PSF convolution)
            therefore the chi^2 function simplifies to
            .. math::
                    \\chi^2 = \\frac{1}{\\sigma^2}(d^TA_{PSF}^{-1}d + \\sum_{i,j}\\alpha_i\\alpha_j x_i^TA_{PSF}x_j - 2\\sum_{i}x_i^Td),
            from which the optimal amplitudes :math:`\\{\\alpha_i\\}` can be solved linearly by solving
            .. math::
                    \\sum_{j} M_{ij}\\alpha_{j} = b_i,
            where :math:`M_{ij} = \\frac{1}{\\sigma^2}x_i^TA_{PSF}x_j` and :math:`b_{i} = \\frac{1}{\\sigma^2}x_i^Td`.

        The steps of this function are:
            (1.) Making the entries :math:`M_{ij}` and :math:`b_i` defined above.
            (2.) Solve the linear function to get the optimal amplitudes.
            (3.) Apply these optimal amplitudes to make unconvolved and convolved model images.
                The output model images are in the form [array1, array2].
                (Note that this is different from the non-interferometric linear solver of Lenstronomy,
                 this output form saves time for likelihood computations in imaging_data for interferometric method.)
                array1 is the unconvolved model image :math:`array1 = \\sum_i \\alpha_i x_i`, where :math:`\\alpha_i` is the solved optimal amplitudes.
                array2 is the convolved model image :math:`array2 = A_{PSF}\\sum_i \\alpha_i x_i`, where :math:`\\alpha_i`.

        :param A: response of unconvolved light profiles, [x_1, x_2, ...]
        :param d: data image, d
        :return: [array1, array2], [amp_array]
        where the [array1, array2] are unconvolved and convolved model images with solved amplitudes
        and [amp_array] are the solved optimal amplitudes.
        """
        num_of_light, num_of_image_pixel = np.shape(A)

        A_convolved = np.zeros(np.shape(A))

        # convolve each response separately
        for i in range(num_of_light):
            A_convolved[i] = util.image2array(
                self._convolution._static_fft(util.array2image(A[i]), mode="same")
            )

        M = np.zeros((num_of_light, num_of_light))
        for i in range(num_of_light):
            for j in range(num_of_light):
                if j < i:
                    M[i, j] = M[j, i]
                else:
                    M[i, j] = np.sum(A_convolved[j] * A[i])

        b = np.zeros((num_of_light))
        for i in range(num_of_light):
            b[i] = np.sum(A[i] * (d))

        param_amps = np.linalg.lstsq(M, b)[0]

        clean_temp = np.zeros((num_of_image_pixel))
        dirty_temp = np.zeros((num_of_image_pixel))
        for i in range(num_of_light):
            clean_temp += param_amps[i] * A[i]
            dirty_temp += param_amps[i] * A_convolved[i]

        clean_model = util.array2image(clean_temp)
        dirty_model = util.array2image(dirty_temp)

        model = [clean_model, dirty_model]

        return model, param_amps



__author__ = "sibirrer"

import pytest
import numpy as np
import copy
import lenstronomy.Util.util as util
import lenstronomy.Util.simulation_util as sim_util
from lenstronomy.ImSim.image_model import ImageModel
from lenstronomy.ImSim.image_linear_solve import ImageLinearFit
import lenstronomy.Util.param_util as param_util
from lenstronomy.PointSource.point_source import PointSource
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.LightModel.light_model import LightModel
from lenstronomy.Workflow.psf_fitting import PsfFitting
from lenstronomy.Data.imaging_data import ImageData
from lenstronomy.Data.psf import PSF


class TestPSFIteration(object):
    """Tests the source model routines."""

    def setup_method(self):
        # data specifics
        sigma_bkg = 0.01  # background noise per pixel
        exp_time = 100  # exposure time (arbitrary units, flux per pixel is in units #photons/exp_time unit)
        numPix = 100  # cutout pixel size
        deltaPix = 0.05  # pixel size in arcsec (area per pixel = deltaPix**2)
        fwhm = 0.3  # full width half max of PSF

        # PSF specification

        kwargs_data = sim_util.data_configure_simple(
            numPix, deltaPix, exp_time, sigma_bkg
        )
        data_class = ImageData(**kwargs_data)
        sigma = util.fwhm2sigma(fwhm)
        x_grid, y_grid = util.make_grid(numPix=31, deltapix=0.05)
        from lenstronomy.LightModel.Profiles.gaussian import Gaussian

        gaussian = Gaussian()
        kernel_point_source = gaussian.function(
            x_grid, y_grid, amp=1.0, sigma=sigma, center_x=0, center_y=0
        )
        kernel_point_source /= np.sum(kernel_point_source)
        kernel_point_source = util.array2image(kernel_point_source)
        psf_error_map = np.zeros_like(kernel_point_source)
        self.kwargs_psf = {
            "psf_type": "PIXEL",
            "kernel_point_source": kernel_point_source,
            "psf_error_map": psf_error_map,
        }

        psf_class = PSF(**self.kwargs_psf)

        # 'EXERNAL_SHEAR': external shear
        kwargs_shear = {
            "gamma1": 0.01,
            "gamma2": 0.01,
        }  # gamma_ext: shear strength, psi_ext: shear angel (in radian)
        phi, q = 0.2, 0.8
        e1, e2 = param_util.phi_q2_ellipticity(phi, q)
        kwargs_spemd = {
            "theta_E": 1.0,
            "gamma": 1.8,
            "center_x": 0,
            "center_y": 0,
            "e1": e1,
            "e2": e2,
        }

        lens_model_list = ["SPEP", "SHEAR"]
        self.kwargs_lens = [kwargs_spemd, kwargs_shear]
        lens_model_class = LensModel(lens_model_list=lens_model_list)
        # list of light profiles (for lens and source)
        # 'SERSIC': spherical Sersic profile
        kwargs_sersic = {
            "amp": 1.0,
            "R_sersic": 0.1,
            "n_sersic": 2,
            "center_x": 0,
            "center_y": 0,
        }
        # 'SERSIC_ELLIPSE': elliptical Sersic profile
        phi, q = 0.2, 0.9
        e1, e2 = param_util.phi_q2_ellipticity(phi, q)
        kwargs_sersic_ellipse = {
            "amp": 1.0,
            "R_sersic": 0.6,
            "n_sersic": 7,
            "center_x": 0,
            "center_y": 0,
            "e1": e1,
            "e2": e2,
        }

        lens_light_model_list = ["SERSIC"]
        self.kwargs_lens_light = [kwargs_sersic]
        lens_light_model_class = LightModel(light_model_list=lens_light_model_list)
        source_model_list = ["SERSIC_ELLIPSE"]
        self.kwargs_source = [kwargs_sersic_ellipse]
        source_model_class = LightModel(light_model_list=source_model_list)
        self.kwargs_ps = [
            {"ra_source": 0.0, "dec_source": 0.0, "source_amp": 10.0}
        ]  # quasar point source position in the source plane and intrinsic brightness
        point_source_class = PointSource(
            point_source_type_list=["SOURCE_POSITION"], fixed_magnification_list=[True]
        )

        kwargs_numerics = {
            "supersampling_factor": 3,
            "supersampling_convolution": False,
            "compute_mode": "regular",
            "point_source_supersampling_factor": 3,
        }
        imageModel = ImageModel(
            data_class,
            psf_class,
            lens_model_class,
            source_model_class,
            lens_light_model_class,
            point_source_class,
            kwargs_numerics=kwargs_numerics,
        )
        image_sim = sim_util.simulate_simple(
            imageModel,
            self.kwargs_lens,
            self.kwargs_source,
            self.kwargs_lens_light,
            self.kwargs_ps,
        )
        data_class.update_data(image_sim)
        self.imageModel = ImageLinearFit(
            data_class,
            psf_class,
            lens_model_class,
            source_model_class,
            lens_light_model_class,
            point_source_class,
            kwargs_numerics=kwargs_numerics,
        )

        self.psf_fitting = PsfFitting(self.imageModel)
        self.kwargs_params = {
            "kwargs_lens": self.kwargs_lens,
            "kwargs_source": self.kwargs_source,
            "kwargs_lens_light": self.kwargs_lens_light,
            "kwargs_ps": self.kwargs_ps,
        }

    def test_update_psf(self):
        fwhm = 0.5
        sigma = util.fwhm2sigma(fwhm)
        x_grid, y_grid = util.make_grid(numPix=31, deltapix=0.05)
        from lenstronomy.LightModel.Profiles.gaussian import Gaussian

        gaussian = Gaussian()
        kernel_point_source = gaussian.function(
            x_grid, y_grid, amp=1.0, sigma=sigma, center_x=0, center_y=0
        )
        kernel_point_source /= np.sum(kernel_point_source)
        kernel_point_source = util.array2image(kernel_point_source)
        kwargs_psf = {"psf_type": "PIXEL", "kernel_point_source": kernel_point_source}

        kwargs_psf_iter = {
            "stacking_method": "median",
            "error_map_radius": 0.5,
            "new_procedure": True,
        }

        kwargs_psf_return, improved_bool, error_map = self.psf_fitting.update_psf(
            kwargs_psf, self.kwargs_params, **kwargs_psf_iter
        )
        assert improved_bool
        kernel_new = kwargs_psf_return["kernel_point_source"]
        kernel_true = self.kwargs_psf["kernel_point_source"]
        kernel_old = kwargs_psf["kernel_point_source"]
        diff_old = np.sum((kernel_old - kernel_true) ** 2)
        diff_new = np.sum((kernel_new - kernel_true) ** 2)
        assert diff_old > diff_new

    def test_calc_corner_mask(self):
        kernel_old = np.ones((101, 101))
        nsymmetry = 4
        corner_mask = self.psf_fitting.calc_cornermask(len(kernel_old), nsymmetry)
        assert corner_mask[corner_mask].size == 0

        nsymmetry = 6
        corner_mask = self.psf_fitting.calc_cornermask(len(kernel_old), nsymmetry)
        assert corner_mask[corner_mask].size < kernel_old.size
        assert corner_mask[corner_mask].size > 0

    def test_combine_psf_corner(self):
        ## start kernel
        kernel_old = np.ones((101, 101))
        test_updated_kernel = copy.deepcopy(kernel_old)
        ##allow the residuals to have different normaliztions
        kernel_list_new = [
            test_updated_kernel * 2,
            test_updated_kernel,
            test_updated_kernel * 4,
            test_updated_kernel,
        ]
        nsymmetry = 6
        corner_mask = self.psf_fitting.calc_cornermask(len(kernel_old), nsymmetry)
        updated_psf = self.psf_fitting.combine_psf(
            kernel_list_new,
            kernel_old,
            factor=1.0,
            stacking_option="median",
            symmetry=nsymmetry,
            corner_symmetry=1,
            corner_mask=corner_mask,
        )
        ##maybe a better criteria here for floats?
        assert abs(updated_psf.max() - updated_psf.min()) < 1e-10
        updated_psf = self.psf_fitting.combine_psf(
            kernel_list_new,
            kernel_old,
            factor=1.0,
            stacking_option="median",
            symmetry=nsymmetry,
            corner_symmetry=2,
            corner_mask=corner_mask,
        )
        assert abs(updated_psf.max() - updated_psf.min()) < 1e-10

    def test_update_iterative(self):
        fwhm = 0.5
        sigma = util.fwhm2sigma(fwhm)
        x_grid, y_grid = util.make_grid(numPix=31, deltapix=0.05)
        from lenstronomy.LightModel.Profiles.gaussian import Gaussian

        gaussian = Gaussian()
        kernel_point_source = gaussian.function(
            x_grid, y_grid, amp=1.0, sigma=sigma, center_x=0, center_y=0
        )
        kernel_point_source /= np.sum(kernel_point_source)
        kernel_point_source = util.array2image(kernel_point_source)
        kwargs_psf = {"psf_type": "PIXEL", "kernel_point_source": kernel_point_source}
        kwargs_psf_iter = {
            "stacking_method": "median",
            "psf_symmetry": 2,
            "psf_iter_factor": 0.2,
            "block_center_neighbour": 0.1,
            "error_map_radius": 0.5,
            "new_procedure": True,
        }

        kwargs_params = copy.deepcopy(self.kwargs_params)
        kwargs_ps = kwargs_params["kwargs_ps"]
        del kwargs_ps[0]["source_amp"]
        print(kwargs_params["kwargs_ps"])
        kwargs_psf_new = self.psf_fitting.update_iterative(
            kwargs_psf, kwargs_params, **kwargs_psf_iter
        )
        kernel_new = kwargs_psf_new["kernel_point_source"]
        kernel_true = self.kwargs_psf["kernel_point_source"]
        kernel_old = kwargs_psf["kernel_point_source"]
        diff_old = np.sum((kernel_old - kernel_true) ** 2)
        diff_new = np.sum((kernel_new - kernel_true) ** 2)
        assert diff_old > diff_new
        assert diff_new < 0.01
        assert "psf_error_map" in kwargs_psf_new

        kwargs_psf_new = self.psf_fitting.update_iterative(
            kwargs_psf,
            kwargs_params,
            num_iter=3,
            no_break=True,
            keep_psf_error_map=True,
        )
        kernel_new = kwargs_psf_new["kernel_point_source"]
        kernel_true = self.kwargs_psf["kernel_point_source"]
        kernel_old = kwargs_psf["kernel_point_source"]
        diff_old = np.sum((kernel_old - kernel_true) ** 2)
        diff_new = np.sum((kernel_new - kernel_true) ** 2)
        assert diff_old > diff_new
        assert diff_new < 0.01

    def test_mask_point_source(self):
        ra_image, dec_image, amp = self.imageModel.PointSource.point_source_list(
            self.kwargs_ps, self.kwargs_lens
        )
        print(ra_image, dec_image, amp)
        x_grid, y_grid = self.imageModel.Data.pixel_coordinates
        x_grid = util.image2array(x_grid)
        y_grid = util.image2array(y_grid)
        radius = 0.5
        mask_point_source = self.psf_fitting.mask_point_source(
            ra_image, dec_image, x_grid, y_grid, radius, i=0
        )
        assert mask_point_source[10, 10] == 1


class TestPSFIterationOld(object):
    """Tests the source model routines."""

    def setup_method(self):
        # data specifics
        sigma_bkg = 0.01  # background noise per pixel
        exp_time = 100  # exposure time (arbitrary units, flux per pixel is in units #photons/exp_time unit)
        numPix = 100  # cutout pixel size
        deltaPix = 0.05  # pixel size in arcsec (area per pixel = deltaPix**2)
        fwhm = 0.3  # full width half max of PSF

        # PSF specification

        kwargs_data = sim_util.data_configure_simple(
            numPix, deltaPix, exp_time, sigma_bkg
        )
        data_class = ImageData(**kwargs_data)
        sigma = util.fwhm2sigma(fwhm)
        x_grid, y_grid = util.make_grid(numPix=31, deltapix=0.05)
        from lenstronomy.LightModel.Profiles.gaussian import Gaussian

        gaussian = Gaussian()
        kernel_point_source = gaussian.function(
            x_grid, y_grid, amp=1.0, sigma=sigma, center_x=0, center_y=0
        )
        kernel_point_source /= np.sum(kernel_point_source)
        kernel_point_source = util.array2image(kernel_point_source)
        psf_error_map = np.zeros_like(kernel_point_source)
        self.kwargs_psf = {
            "psf_type": "PIXEL",
            "kernel_point_source": kernel_point_source,
            "psf_error_map": psf_error_map,
        }

        psf_class = PSF(**self.kwargs_psf)

        # 'EXERNAL_SHEAR': external shear
        kwargs_shear = {
            "gamma1": 0.01,
            "gamma2": 0.01,
        }  # gamma_ext: shear strength, psi_ext: shear angel (in radian)
        phi, q = 0.2, 0.8
        e1, e2 = param_util.phi_q2_ellipticity(phi, q)
        kwargs_spemd = {
            "theta_E": 1.0,
            "gamma": 1.8,
            "center_x": 0,
            "center_y": 0,
            "e1": e1,
            "e2": e2,
        }

        lens_model_list = ["SPEP", "SHEAR"]
        self.kwargs_lens = [kwargs_spemd, kwargs_shear]
        lens_model_class = LensModel(lens_model_list=lens_model_list)
        # list of light profiles (for lens and source)
        # 'SERSIC': spherical Sersic profile
        kwargs_sersic = {
            "amp": 1.0,
            "R_sersic": 0.1,
            "n_sersic": 2,
            "center_x": 0,
            "center_y": 0,
        }
        # 'SERSIC_ELLIPSE': elliptical Sersic profile
        phi, q = 0.2, 0.9
        e1, e2 = param_util.phi_q2_ellipticity(phi, q)
        kwargs_sersic_ellipse = {
            "amp": 1.0,
            "R_sersic": 0.6,
            "n_sersic": 7,
            "center_x": 0,
            "center_y": 0,
            "e1": e1,
            "e2": e2,
        }

        lens_light_model_list = ["SERSIC"]
        self.kwargs_lens_light = [kwargs_sersic]
        lens_light_model_class = LightModel(light_model_list=lens_light_model_list)
        source_model_list = ["SERSIC_ELLIPSE"]
        self.kwargs_source = [kwargs_sersic_ellipse]
        source_model_class = LightModel(light_model_list=source_model_list)
        self.kwargs_ps = [
            {"ra_source": 0.0, "dec_source": 0.0, "source_amp": 10.0}
        ]  # quasar point source position in the source plane and intrinsic brightness
        point_source_class = PointSource(
            point_source_type_list=["SOURCE_POSITION"], fixed_magnification_list=[True]
        )

        kwargs_numerics = {
            "supersampling_factor": 3,
            "supersampling_convolution": False,
            "compute_mode": "regular",
            "point_source_supersampling_factor": 3,
        }
        imageModel = ImageModel(
            data_class,
            psf_class,
            lens_model_class,
            source_model_class,
            lens_light_model_class,
            point_source_class,
            kwargs_numerics=kwargs_numerics,
        )
        image_sim = sim_util.simulate_simple(
            imageModel,
            self.kwargs_lens,
            self.kwargs_source,
            self.kwargs_lens_light,
            self.kwargs_ps,
        )
        data_class.update_data(image_sim)
        self.imageModel = ImageLinearFit(
            data_class,
            psf_class,
            lens_model_class,
            source_model_class,
            lens_light_model_class,
            point_source_class,
            kwargs_numerics=kwargs_numerics,
        )

        self.psf_fitting = PsfFitting(self.imageModel)
        self.kwargs_params = {
            "kwargs_lens": self.kwargs_lens,
            "kwargs_source": self.kwargs_source,
            "kwargs_lens_light": self.kwargs_lens_light,
            "kwargs_ps": self.kwargs_ps,
        }

    def test_update_psf(self):
        fwhm = 0.5
        sigma = util.fwhm2sigma(fwhm)
        x_grid, y_grid = util.make_grid(numPix=31, deltapix=0.05)
        from lenstronomy.LightModel.Profiles.gaussian import Gaussian

        gaussian = Gaussian()
        kernel_point_source = gaussian.function(
            x_grid, y_grid, amp=1.0, sigma=sigma, center_x=0, center_y=0
        )
        kernel_point_source /= np.sum(kernel_point_source)
        kernel_point_source = util.array2image(kernel_point_source)
        kwargs_psf = {"psf_type": "PIXEL", "kernel_point_source": kernel_point_source}

        kwargs_psf_iter = {
            "stacking_method": "median",
            "error_map_radius": 0.5,
            "new_procedure": False,
        }

        kwargs_psf_return, improved_bool, error_map = self.psf_fitting.update_psf(
            kwargs_psf, self.kwargs_params, **kwargs_psf_iter
        )
        assert improved_bool
        kernel_new = kwargs_psf_return["kernel_point_source"]
        kernel_true = self.kwargs_psf["kernel_point_source"]
        kernel_old = kwargs_psf["kernel_point_source"]
        diff_old = np.sum((kernel_old - kernel_true) ** 2)
        diff_new = np.sum((kernel_new - kernel_true) ** 2)
        assert diff_old > diff_new

    def test_update_iterative(self):
        fwhm = 0.5
        sigma = util.fwhm2sigma(fwhm)
        x_grid, y_grid = util.make_grid(numPix=31, deltapix=0.05)
        from lenstronomy.LightModel.Profiles.gaussian import Gaussian

        gaussian = Gaussian()
        kernel_point_source = gaussian.function(
            x_grid, y_grid, amp=1.0, sigma=sigma, center_x=0, center_y=0
        )
        kernel_point_source /= np.sum(kernel_point_source)
        kernel_point_source = util.array2image(kernel_point_source)
        kwargs_psf = {
            "psf_type": "PIXEL",
            "kernel_point_source": kernel_point_source,
            "kernel_point_source_init": kernel_point_source,
        }
        kwargs_psf_iter = {
            "stacking_method": "median",
            "psf_symmetry": 2,
            "psf_iter_factor": 0.2,
            "block_center_neighbour": 0.1,
            "error_map_radius": 0.5,
            "new_procedure": False,
            "no_break": False,
            "verbose": True,
            "keep_psf_error_map": False,
        }

        kwargs_params = copy.deepcopy(self.kwargs_params)
        kwargs_ps = kwargs_params["kwargs_ps"]
        del kwargs_ps[0]["source_amp"]
        print(kwargs_params["kwargs_ps"])
        kwargs_psf_new = self.psf_fitting.update_iterative(
            kwargs_psf, kwargs_params, **kwargs_psf_iter
        )
        kernel_new = kwargs_psf_new["kernel_point_source"]
        kernel_true = self.kwargs_psf["kernel_point_source"]
        kernel_old = kwargs_psf["kernel_point_source"]
        diff_old = np.sum((kernel_old - kernel_true) ** 2)
        diff_new = np.sum((kernel_new - kernel_true) ** 2)
        assert diff_old > diff_new
        assert diff_new < 0.01
        assert "psf_error_map" in kwargs_psf_new

        kwargs_psf_new = self.psf_fitting.update_iterative(
            kwargs_psf,
            kwargs_params,
            num_iter=3,
            no_break=True,
            keep_psf_error_map=True,
        )
        kernel_new = kwargs_psf_new["kernel_point_source"]
        kernel_true = self.kwargs_psf["kernel_point_source"]
        kernel_old = kwargs_psf["kernel_point_source"]
        diff_old = np.sum((kernel_old - kernel_true) ** 2)
        diff_new = np.sum((kernel_new - kernel_true) ** 2)
        assert diff_old > diff_new
        assert diff_new < 0.01


if __name__ == "__main__":
    pytest.main()



__author__ = "sibirrer"

import unittest
from lenstronomy.LensModel.Profiles.hessian import Hessian
from lenstronomy.LensModel.lens_model import LensModel

import numpy as np
import numpy.testing as npt
import pytest


class TestHessian(object):
    """Tests the Gaussian methods."""

    def setup_method(self):
        self.hessian = Hessian()

        self.f_xx, self.f_yy, self.f_xy, self.f_yx = 0.1, 0.1, -0.1, -0.1
        self.kwargs_lens = {
            "f_xx": self.f_xx,
            "f_yy": self.f_yy,
            "f_xy": self.f_xy,
            "f_yx": self.f_yx,
        }

    def test_function(self):
        x = 1
        y = 2
        values = self.hessian.function(x, y, **self.kwargs_lens)
        f_true = (
            1
            / 2.0
            * (
                self.f_xx * x**2
                + (self.f_xy + self.f_yx) * x * y
                + self.f_yy * y**2
            )
        )
        npt.assert_almost_equal(values, f_true, decimal=5)
        x = np.array([0])
        y = np.array([0])
        values = self.hessian.function(x, y, **self.kwargs_lens)
        npt.assert_almost_equal(values[0], 0, decimal=5)

    def test_derivatives(self):
        x = 1
        y = 2
        f_x, f_y = self.hessian.derivatives(x, y, **self.kwargs_lens)
        fx_true = self.f_xx * x + self.f_xy * y
        fy_true = self.f_yx * x + self.f_yy * y
        npt.assert_almost_equal(f_x, fx_true, decimal=5)
        npt.assert_almost_equal(f_y, fy_true, decimal=5)

    def test_hessian(self):
        x = 1
        y = 2

        f_xx, f_xy, f_yx, f_yy = self.hessian.hessian(x, y, **self.kwargs_lens)
        npt.assert_almost_equal(f_xx, self.f_xx, decimal=5)
        npt.assert_almost_equal(f_yy, self.f_yy, decimal=5)
        npt.assert_almost_equal(f_xy, self.f_xy, decimal=5)
        npt.assert_almost_equal(f_yx, self.f_yx, decimal=5)

        lensModel = LensModel(["HESSIAN"])
        f_xy_true, f_yx_true = 0.3, 0.2
        kwargs_lens = {
            "f_xx": self.f_xx,
            "f_yy": self.f_yy,
            "f_xy": f_xy_true,
            "f_yx": f_yx_true,
        }
        f_xx, f_xy, f_yx, f_yy = lensModel.hessian(x, y, [kwargs_lens], diff=0.001)
        npt.assert_almost_equal(f_xx, self.f_xx, decimal=9)
        npt.assert_almost_equal(f_yy, self.f_yy, decimal=9)
        npt.assert_almost_equal(f_xy, f_xy_true, decimal=9)
        npt.assert_almost_equal(f_yx, f_yx_true, decimal=9)


if __name__ == "__main__":
    pytest.main()



"""Routines to compute derivatives of spherical functions."""
import numpy as np

from lenstronomy.Util.package_util import exporter

export, __all__ = exporter()


@export
def d_r_dx(x, y):
    """Derivative of r with respect to x :param x:

    :param y:
    :return:
    """
    return x / np.sqrt(x**2 + y**2)


@export
def d_r_dy(x, y):
    """Differential dr/dy.

    :param x:
    :param y:
    :return:
    """
    return y / np.sqrt(x**2 + y**2)


@export
def d_r_dxx(x, y):
    """Second derivative dr/dxdx :param x:

    :param y:
    :return:
    """
    return y**2 / (x**2 + y**2) ** (3.0 / 2)


@export
def d_r_dyy(x, y):
    """Second derivative dr/dxdx :param x:

    :param y:
    :return:
    """
    return x**2 / (x**2 + y**2) ** (3.0 / 2)


@export
def d_r_dxy(x, y):
    """Second derivative dr/dxdx :param x:

    :param y:
    :return:
    """
    return -x * y / (x**2 + y**2) ** (3 / 2.0)


@export
def d_phi_dx(x, y):
    """Angular derivative in respect to x when phi = arctan2(y, x)

    :param x:
    :param y:
    :return:
    """
    return -y / (x**2 + y**2)


@export
def d_phi_dy(x, y):
    """Angular derivative in respect to y when phi = arctan2(y, x)

    :param x:
    :param y:
    :return:
    """
    return x / (x**2 + y**2)


@export
def d_phi_dxx(x, y):
    """Second derivative of the orientation angle.

    :param x:
    :param y:
    :return:
    """
    return 2 * x * y / (x**2 + y**2) ** 2


@export
def d_phi_dyy(x, y):
    """Second derivative of the orientation angle in dydy.

    :param x:
    :param y:
    :return:
    """
    return -2 * x * y / (x**2 + y**2) ** 2


@export
def d_phi_dxy(x, y):
    """Second derivative of the orientation angle in dxdy.

    :param x:
    :param y:
    :return:
    """
    return (-(x**2) + y**2) / (x**2 + y**2) ** 2


@export
def d_x_diffr_dx(x, y):
    """Derivative of d(x/r)/dx equivalent to second order derivatives dr_dxx.

    :param x:
    :param y:
    :return:
    """
    return y**2 / (x**2 + y**2) ** (3 / 2.0)


@export
def d_y_diffr_dy(x, y):
    """Derivative of d(y/r)/dy equivalent to second order derivatives dr_dyy.

    :param x:
    :param y:
    :return:
    """
    return x**2 / (x**2 + y**2) ** (3 / 2.0)


@export
def d_y_diffr_dx(x, y):
    """Derivative of d(y/r)/dx equivalent to second order derivatives dr_dxy.

    :param x:
    :param y:
    :return:
    """
    return -x * y / (x**2 + y**2) ** (3 / 2.0)


@export
def d_x_diffr_dy(x, y):
    """Derivative of d(x/r)/dy equivalent to second order derivatives dr_dyx.

    :param x:
    :param y:
    :return:
    """
    return -x * y / (x**2 + y**2) ** (3 / 2.0)



"""Tests for `galkin` module."""
import pytest
import unittest
import copy
import numpy.testing as npt
import numpy as np
import scipy.integrate as integrate
from lenstronomy.GalKin.galkin import Galkin
from lenstronomy.GalKin.light_profile import LightProfile
import lenstronomy.Util.param_util as param_util
from lenstronomy.Util import constants as const


class TestGalkin(object):
    def setup_method(self):
        np.random.seed(42)

    def test_compare_power_law(self):
        """Compare power-law profiles analytical vs.

        numerical
        :return:
        """
        # light profile
        light_profile_list = ["HERNQUIST"]
        r_eff = 1.5
        kwargs_light = [
            {"Rs": 0.551 * r_eff, "amp": 1.0}
        ]  # effective half light radius (2d projected) in arcsec
        # 0.551 *
        # mass profile
        mass_profile_list = ["SPP"]
        theta_E = 1.2
        gamma = 2.0
        kwargs_profile = [
            {"theta_E": theta_E, "gamma": gamma}
        ]  # Einstein radius (arcsec) and power-law slope

        # anisotropy profile
        anisotropy_type = "OM"
        r_ani = 2.0
        kwargs_anisotropy = {"r_ani": r_ani}  # anisotropy radius [arcsec]

        # aperture as slit
        aperture_type = "slit"
        length = 1.0
        width = 0.3
        kwargs_aperture = {
            "aperture_type": aperture_type,
            "length": length,
            "width": width,
            "center_ra": 0,
            "center_dec": 0,
            "angle": 0,
        }

        psf_fwhm = 1.0  # Gaussian FWHM psf
        kwargs_cosmo = {"d_d": 1000, "d_s": 1500, "d_ds": 800}
        kwargs_numerics = {
            "interpol_grid_num": 1000,
            "max_integrate": 1000,
            "min_integrate": 0.001,
        }
        kwargs_model = {
            "mass_profile_list": mass_profile_list,
            "light_profile_list": light_profile_list,
            "anisotropy_model": anisotropy_type,
        }
        kwargs_psf = {"psf_type": "GAUSSIAN", "fwhm": psf_fwhm}

        galkin_analytic = Galkin(
            kwargs_model=kwargs_model,
            kwargs_aperture=kwargs_aperture,
            kwargs_psf=kwargs_psf,
            kwargs_cosmo=kwargs_cosmo,
            kwargs_numerics=kwargs_numerics,
            analytic_kinematics=True,
        )
        sigma_v_analytic = galkin_analytic.dispersion(
            kwargs_mass={"gamma": gamma, "theta_E": theta_E},
            kwargs_light={"r_eff": r_eff},
            kwargs_anisotropy={"r_ani": r_ani},
            sampling_number=1000,
        )
        kwargs_numerics["lum_weight_int_method"] = False
        galkin_num_3d = Galkin(
            kwargs_model=kwargs_model,
            kwargs_aperture=kwargs_aperture,
            kwargs_psf=kwargs_psf,
            kwargs_cosmo=kwargs_cosmo,
            kwargs_numerics=kwargs_numerics,
            analytic_kinematics=False,
        )
        sigma_v_num_3d = galkin_num_3d.dispersion(
            kwargs_profile, kwargs_light, kwargs_anisotropy, sampling_number=1000
        )

        print(
            sigma_v_analytic,
            sigma_v_num_3d,
            "sigma_v Galkin 3d numerics, sigma_v analytic",
        )
        npt.assert_almost_equal(sigma_v_num_3d / sigma_v_analytic, 1, decimal=2)

        # 2d projected integral calculation
        kwargs_numerics = {
            "interpol_grid_num": 1000,
            "max_integrate": 1000,
            "min_integrate": 0.000001,
            "lum_weight_int_method": True,
            "log_integration": True,
        }
        galkin_num_log_proj = Galkin(
            kwargs_model=kwargs_model,
            kwargs_aperture=kwargs_aperture,
            kwargs_psf=kwargs_psf,
            kwargs_cosmo=kwargs_cosmo,
            kwargs_numerics=kwargs_numerics,
            analytic_kinematics=False,
        )
        sigma_v_num_log_proj = galkin_num_log_proj.dispersion(
            kwargs_profile, kwargs_light, kwargs_anisotropy, sampling_number=1000
        )

        kwargs_numerics = {
            "interpol_grid_num": 10000,
            "max_integrate": 1000,
            "min_integrate": 0.0001,
            "lum_weight_int_method": True,
            "log_integration": False,
        }
        galkin_num_lin_proj = Galkin(
            kwargs_model=kwargs_model,
            kwargs_aperture=kwargs_aperture,
            kwargs_psf=kwargs_psf,
            kwargs_cosmo=kwargs_cosmo,
            kwargs_numerics=kwargs_numerics,
            analytic_kinematics=False,
        )
        sigma_v_num_lin_proj = galkin_num_lin_proj.dispersion(
            kwargs_profile, kwargs_light, kwargs_anisotropy, sampling_number=1000
        )

        print(
            sigma_v_num_log_proj / sigma_v_analytic,
            sigma_v_num_lin_proj / sigma_v_analytic,
            "log proj, lin proj",
        )

        npt.assert_almost_equal(sigma_v_num_log_proj / sigma_v_analytic, 1, decimal=2)
        npt.assert_almost_equal(sigma_v_num_lin_proj / sigma_v_analytic, 1, decimal=2)

    def test_log_vs_linear_integral(self):
        """Here we test logarithmic vs linear integral in an end-to-end fashion.

        We do not demand the highest level of precisions here!!! We are using the
        luminosity-weighted velocity dispersion integration calculation in this test.
        """

        # light profile
        light_profile_list = ["HERNQUIST"]
        Rs = 0.5
        kwargs_light = [
            {"Rs": Rs, "amp": 1.0}
        ]  # effective half light radius (2d projected) in arcsec
        # 0.551 *
        # mass profile
        mass_profile_list = ["SPP"]
        theta_E = 1.2
        gamma = 2.0
        kwargs_profile = [
            {"theta_E": theta_E, "gamma": gamma}
        ]  # Einstein radius (arcsec) and power-law slope

        # anisotropy profile
        anisotropy_type = "OM"
        r_ani = 2.0
        kwargs_anisotropy = {"r_ani": r_ani}  # anisotropy radius [arcsec]

        # aperture as slit
        aperture_type = "slit"
        length = 3.8
        width = 0.9
        kwargs_aperture = {
            "aperture_type": aperture_type,
            "length": length,
            "width": width,
            "center_ra": 0,
            "center_dec": 0,
            "angle": 0,
        }

        psf_fwhm = 0.7  # Gaussian FWHM psf
        kwargs_cosmo = {"d_d": 1000, "d_s": 1500, "d_ds": 800}
        kwargs_numerics_log = {
            "interpol_grid_num": 1000,
            "log_integration": True,
            "max_integrate": 10,
            "min_integrate": 0.001,
            "lum_weight_int_method": True,
        }
        kwargs_numerics_linear = {
            "interpol_grid_num": 1000,
            "log_integration": False,
            "max_integrate": 10,
            "min_integrate": 0.001,
            "lum_weight_int_method": True,
        }
        kwargs_psf = {"psf_type": "GAUSSIAN", "fwhm": psf_fwhm}
        kwargs_model = {
            "mass_profile_list": mass_profile_list,
            "light_profile_list": light_profile_list,
            "anisotropy_model": anisotropy_type,
        }
        galkin_linear = Galkin(
            kwargs_model=kwargs_model,
            kwargs_aperture=kwargs_aperture,
            kwargs_psf=kwargs_psf,
            kwargs_cosmo=kwargs_cosmo,
            kwargs_numerics=kwargs_numerics_linear,
        )

        sigma_v_lin = galkin_linear.dispersion(
            kwargs_profile, kwargs_light, kwargs_anisotropy, sampling_number=1000
        )
        galkin_log = Galkin(
            kwargs_model=kwargs_model,
            kwargs_aperture=kwargs_aperture,
            kwargs_psf=kwargs_psf,
            kwargs_cosmo=kwargs_cosmo,
            kwargs_numerics=kwargs_numerics_log,
        )
        sigma_v_log = galkin_log.dispersion(
            kwargs_profile, kwargs_light, kwargs_anisotropy, sampling_number=1000
        )
        print(sigma_v_lin, sigma_v_log, "sigma_v linear, sigma_v log")
        print((sigma_v_lin / sigma_v_log) ** 2)

        npt.assert_almost_equal(sigma_v_lin / sigma_v_log, 1, decimal=2)

    def test_projected_light_integral_hernquist(self):
        """

        :return:
        """
        light_profile_list = ["HERNQUIST"]
        Rs = 1.0
        kwargs_light = [
            {"Rs": Rs, "amp": 1.0}
        ]  # effective half light radius (2d projected) in arcsec
        lightProfile = LightProfile(light_profile_list)
        R = 2
        light2d = lightProfile.light_2d(R=R, kwargs_list=kwargs_light)
        out = integrate.quad(
            lambda x: lightProfile.light_3d(np.sqrt(R**2 + x**2), kwargs_light),
            0,
            100,
        )
        npt.assert_almost_equal(light2d, out[0] * 2, decimal=3)

    def test_projected_light_integral_hernquist_ellipse(self):
        """

        :return:
        """
        light_profile_list = ["HERNQUIST_ELLIPSE"]
        Rs = 1.0
        phi, q = 1, 0.8
        e1, e2 = param_util.phi_q2_ellipticity(phi, q)
        kwargs_light = [
            {"Rs": Rs, "amp": 1.0, "e1": e1, "e2": e2}
        ]  # effective half light radius (2d projected) in arcsec
        lightProfile = LightProfile(light_profile_list)
        R = 2
        light2d = lightProfile.light_2d(R=R, kwargs_list=kwargs_light)
        out = integrate.quad(
            lambda x: lightProfile.light_3d(np.sqrt(R**2 + x**2), kwargs_light),
            0,
            10,
        )
        npt.assert_almost_equal(light2d, out[0] * 2, decimal=3)

    def test_projected_light_integral_pjaffe(self):
        """

        :return:
        """
        light_profile_list = ["PJAFFE"]
        kwargs_light = [
            {"Rs": 0.5, "Ra": 0.01, "amp": 1.0}
        ]  # effective half light radius (2d projected) in arcsec
        lightProfile = LightProfile(light_profile_list)
        R = 0.01
        light2d = lightProfile.light_2d(R=R, kwargs_list=kwargs_light)
        out = integrate.quad(
            lambda x: lightProfile.light_3d(np.sqrt(R**2 + x**2), kwargs_light),
            0,
            100,
        )
        print(out, "out")
        npt.assert_almost_equal(light2d / (out[0] * 2), 1.0, decimal=3)

    def test_realistic_0(self):
        """Realistic test example :return:"""
        light_profile_list = ["HERNQUIST"]
        kwargs_light = [
            {
                "Rs": 0.10535462602138289,
                "center_x": -0.02678473951679429,
                "center_y": 0.88691126347462712,
                "amp": 3.7114695634960109,
            }
        ]
        lightProfile = LightProfile(light_profile_list)
        R = 0.01
        light2d = lightProfile.light_2d(R=R, kwargs_list=kwargs_light)
        out = integrate.quad(
            lambda x: lightProfile.light_3d(np.sqrt(R**2 + x**2), kwargs_light),
            0,
            100,
        )
        print(out, "out")
        npt.assert_almost_equal(light2d / (out[0] * 2), 1.0, decimal=3)

    def test_realistic_1(self):
        """Realistic test example :return:"""
        light_profile_list = ["HERNQUIST_ELLIPSE"]
        phi, q = 0.74260706384506325, 0.46728323131925864
        e1, e2 = param_util.phi_q2_ellipticity(phi, q)
        kwargs_light = [
            {
                "Rs": 0.10535462602138289,
                "e1": e1,
                "e2": e2,
                "center_x": -0.02678473951679429,
                "center_y": 0.88691126347462712,
                "amp": 3.7114695634960109,
            }
        ]
        lightProfile = LightProfile(light_profile_list)
        R = 0.01
        light2d = lightProfile.light_2d(R=R, kwargs_list=kwargs_light)
        out = integrate.quad(
            lambda x: lightProfile.light_3d(np.sqrt(R**2 + x**2), kwargs_light),
            0,
            100,
        )
        print(out, "out")
        npt.assert_almost_equal(light2d / (out[0] * 2), 1.0, decimal=3)

    def test_realistic(self):
        """Realistic test example :return:"""
        light_profile_list = ["HERNQUIST_ELLIPSE", "PJAFFE_ELLIPSE"]
        phi, q = 0.74260706384506325, 0.46728323131925864
        e1, e2 = param_util.phi_q2_ellipticity(phi, q)

        phi2, q2 = -0.33379268413794494, 0.66582356813012267
        e12, e22 = param_util.phi_q2_ellipticity(phi2, q2)
        kwargs_light = [
            {
                "Rs": 0.10535462602138289,
                "e1": e1,
                "e2": e2,
                "center_x": -0.02678473951679429,
                "center_y": 0.88691126347462712,
                "amp": 3.7114695634960109,
            },
            {
                "Rs": 0.44955054610388684,
                "e1": e12,
                "e2": e22,
                "center_x": 0.019536801118136753,
                "center_y": 0.0218888643537157,
                "Ra": 0.0010000053334891974,
                "amp": 967.00280526319796,
            },
        ]
        lightProfile = LightProfile(light_profile_list)
        R = 0.01
        light2d = lightProfile.light_2d(R=R, kwargs_list=kwargs_light)
        out = integrate.quad(
            lambda x: lightProfile.light_3d(np.sqrt(R**2 + x**2), kwargs_light),
            0,
            100,
        )
        print(out, "out")
        npt.assert_almost_equal(light2d / (out[0] * 2), 1.0, decimal=3)

    def test_dispersion_map(self):
        """Tests whether the old and new version provide the same answer."""
        # light profile
        light_profile_list = ["HERNQUIST"]
        r_eff = 1.5
        kwargs_light = [
            {"Rs": r_eff, "amp": 1.0}
        ]  # effective half light radius (2d projected) in arcsec
        # 0.551 *
        # mass profile
        mass_profile_list = ["SPP"]
        theta_E = 1.2
        gamma = 2.0
        kwargs_mass = [
            {"theta_E": theta_E, "gamma": gamma}
        ]  # Einstein radius (arcsec) and power-law slope

        # anisotropy profile
        anisotropy_type = "OM"
        r_ani = 2.0
        kwargs_anisotropy = {"r_ani": r_ani}  # anisotropy radius [arcsec]

        # aperture as shell
        # aperture_type = 'shell'
        # kwargs_aperture_inner = {'r_in': 0., 'r_out': 0.2, 'center_dec': 0, 'center_ra': 0}

        # kwargs_aperture_outer = {'r_in': 0., 'r_out': 1.5, 'center_dec': 0, 'center_ra': 0}

        # aperture as slit
        r_bins = np.linspace(0, 2, 3)
        kwargs_ifu = {
            "r_bins": r_bins,
            "center_ra": 0,
            "center_dec": 0,
            "aperture_type": "IFU_shells",
        }
        kwargs_aperture = {
            "aperture_type": "shell",
            "r_in": r_bins[0],
            "r_out": r_bins[1],
            "center_ra": 0,
            "center_dec": 0,
        }

        psf_fwhm = 1.0  # Gaussian FWHM psf
        kwargs_cosmo = {"d_d": 1000, "d_s": 1500, "d_ds": 800}
        kwargs_numerics = {
            "interpol_grid_num": 500,
            "log_integration": True,
            "max_integrate": 100,
        }
        kwargs_model = {
            "mass_profile_list": mass_profile_list,
            "light_profile_list": light_profile_list,
            "anisotropy_model": anisotropy_type,
        }
        kwargs_psf = {"psf_type": "GAUSSIAN", "fwhm": psf_fwhm}

        galkinIFU = Galkin(
            kwargs_aperture=kwargs_ifu,
            kwargs_psf=kwargs_psf,
            kwargs_cosmo=kwargs_cosmo,
            kwargs_model=kwargs_model,
            kwargs_numerics=kwargs_numerics,
            analytic_kinematics=True,
        )
        sigma_v_ifu = galkinIFU.dispersion_map(
            kwargs_mass={"theta_E": theta_E, "gamma": gamma},
            kwargs_light={"r_eff": r_eff},
            kwargs_anisotropy=kwargs_anisotropy,
            num_kin_sampling=1000,
        )
        galkin = Galkin(
            kwargs_model,
            kwargs_aperture,
            kwargs_psf,
            kwargs_cosmo,
            kwargs_numerics,
            analytic_kinematics=True,
        )
        sigma_v = galkin.dispersion(
            kwargs_mass={"theta_E": theta_E, "gamma": gamma},
            kwargs_light={"r_eff": r_eff},
            kwargs_anisotropy=kwargs_anisotropy,
            sampling_number=1000,
        )
        npt.assert_almost_equal(sigma_v, sigma_v_ifu[0], decimal=-1)

    def test_projected_integral_vs_3d_rendering(self):
        lum_weight_int_method = True

        # light profile
        light_profile_list = ["HERNQUIST"]
        r_eff = 1.5
        kwargs_light = [
            {"Rs": 0.551 * r_eff, "amp": 1.0}
        ]  # effective half light radius (2d projected) in arcsec
        # 0.551 *
        # mass profile
        mass_profile_list = ["SPP"]
        theta_E = 1.2
        gamma = 2.0
        kwargs_profile = [
            {"theta_E": theta_E, "gamma": gamma}
        ]  # Einstein radius (arcsec) and power-law slope

        # anisotropy profile
        anisotropy_type = "OM"
        r_ani = 2.0
        kwargs_anisotropy = {"r_ani": r_ani}  # anisotropy radius [arcsec]

        # aperture as slit
        aperture_type = "slit"
        length = 1.0
        width = 0.3
        kwargs_aperture = {
            "aperture_type": aperture_type,
            "length": length,
            "width": width,
            "center_ra": 0,
            "center_dec": 0,
            "angle": 0,
        }

        psf_fwhm = 1.0  # Gaussian FWHM psf
        kwargs_cosmo = {"d_d": 1000, "d_s": 1500, "d_ds": 800}
        kwargs_numerics_3d = {
            "interpol_grid_num": 2000,
            "log_integration": True,
            "max_integrate": 1000,
            "min_integrate": 0.00001,
            "lum_weight_int_method": False,
        }
        kwargs_model = {
            "mass_profile_list": mass_profile_list,
            "light_profile_list": light_profile_list,
            "anisotropy_model": anisotropy_type,
        }
        kwargs_psf = {"psf_type": "GAUSSIAN", "fwhm": psf_fwhm}
        galkin = Galkin(
            kwargs_model=kwargs_model,
            kwargs_aperture=kwargs_aperture,
            kwargs_psf=kwargs_psf,
            kwargs_cosmo=kwargs_cosmo,
            kwargs_numerics=kwargs_numerics_3d,
        )
        sigma_v = galkin.dispersion(
            kwargs_profile, kwargs_light, kwargs_anisotropy, sampling_number=1000
        )

        kwargs_numerics_2d = {
            "interpol_grid_num": 2000,
            "log_integration": True,
            "max_integrate": 1000,
            "min_integrate": 0.00001,
            "lum_weight_int_method": True,
        }

        galkin = Galkin(
            kwargs_model=kwargs_model,
            kwargs_aperture=kwargs_aperture,
            kwargs_psf=kwargs_psf,
            kwargs_cosmo=kwargs_cosmo,
            kwargs_numerics=kwargs_numerics_2d,
            analytic_kinematics=False,
        )
        sigma_v_int_method = galkin.dispersion(
            kwargs_profile, kwargs_light, kwargs_anisotropy, sampling_number=1000
        )
        npt.assert_almost_equal(sigma_v_int_method / sigma_v, 1, decimal=2)

    def test_2d_vs_3d_power_law(self):
        # set up power-law light profile
        light_model = ["POWER_LAW"]
        kwargs_light = [{"gamma": 2, "amp": 1, "e1": 0, "e2": 0}]

        lens_model = ["SIS"]
        kwargs_mass = [{"theta_E": 1}]

        anisotropy_type = "isotropic"
        kwargs_anisotropy = {}
        kwargs_model = {
            "mass_profile_list": lens_model,
            "light_profile_list": light_model,
            "anisotropy_model": anisotropy_type,
        }
        kwargs_numerics = {
            "interpol_grid_num": 2000,
            "log_integration": True,
            "max_integrate": 50,
            "min_integrate": 0.0001,
        }

        kwargs_numerics_3d = copy.deepcopy(kwargs_numerics)
        kwargs_numerics_3d["lum_weight_int_method"] = False

        kwargs_numerics_2d = copy.deepcopy(kwargs_numerics)
        kwargs_numerics_2d["lum_weight_int_method"] = True

        kwargs_cosmo = {"d_d": 1000, "d_s": 1500, "d_ds": 800}

        # compute analytic velocity dispersion of SIS profile

        v_sigma_c2 = (
            kwargs_mass[0]["theta_E"]
            * const.arcsec
            / (4 * np.pi)
            * kwargs_cosmo["d_s"]
            / kwargs_cosmo["d_ds"]
        )
        v_sigma_true = np.sqrt(v_sigma_c2) * const.c / 1000

        # aperture as slit
        aperture_type = "slit"
        length = 1.0
        width = 0.3
        kwargs_aperture = {
            "aperture_type": aperture_type,
            "length": length,
            "width": width,
            "center_ra": 0,
            "center_dec": 0,
            "angle": 0,
        }
        kwargs_psf = {"psf_type": "GAUSSIAN", "fwhm": 0.5}

        galkin3d = Galkin(
            kwargs_model=kwargs_model,
            kwargs_aperture=kwargs_aperture,
            kwargs_psf=kwargs_psf,
            kwargs_cosmo=kwargs_cosmo,
            kwargs_numerics=kwargs_numerics_3d,
        )

        galkin2d = Galkin(
            kwargs_model=kwargs_model,
            kwargs_aperture=kwargs_aperture,
            kwargs_psf=kwargs_psf,
            kwargs_cosmo=kwargs_cosmo,
            kwargs_numerics=kwargs_numerics_2d,
        )

        sigma_draw_list = []
        for i in range(100):
            sigma_v_draw = galkin3d._draw_one_sigma2(
                kwargs_mass, kwargs_light, kwargs_anisotropy
            )
            sigma_draw_list.append(sigma_v_draw)
            # print(np.sqrt(sigma_v_draw)/ 1000)

        # import matplotlib.pyplot as plt
        # plt.plot(np.sqrt(sigma_draw_list) / 1000 / v_sigma_true)
        # plt.show()

        print(np.sqrt(np.mean(sigma_draw_list)) / 1000, "mean draw")
        print("truth = ", v_sigma_true)
        # assert 1 == 0

        sigma_v_2d = galkin2d.dispersion(
            kwargs_mass, kwargs_light, kwargs_anisotropy, sampling_number=1000
        )
        sigma_v_3d = galkin3d.dispersion(
            kwargs_mass, kwargs_light, kwargs_anisotropy, sampling_number=1000
        )
        npt.assert_almost_equal(sigma_v_2d / v_sigma_true, 1, decimal=2)
        npt.assert_almost_equal(sigma_v_3d / v_sigma_true, 1, decimal=2)


class TestRaise(unittest.TestCase):
    def test_raise(self):
        with self.assertRaises(ValueError):
            kwargs_model = {"anisotropy_model": "const"}
            kwargs_aperture = {
                "center_ra": 0,
                "width": 1,
                "length": 1,
                "angle": 0,
                "center_dec": 0,
                "aperture_type": "slit",
            }
            kwargs_cosmo = {"d_d": 1000, "d_s": 1500, "d_ds": 800}
            kwargs_psf = {"psf_type": "GAUSSIAN", "fwhm": 1}
            Galkin(
                kwargs_model,
                kwargs_aperture,
                kwargs_psf,
                kwargs_cosmo,
                kwargs_numerics={},
                analytic_kinematics=True,
            )


if __name__ == "__main__":
    pytest.main()



import copy
import numpy as np
import pytest

from numpy.testing import assert_raises

from lenstronomy.Sampling.Samplers.nautilus_sampler import NautilusSampler


@pytest.fixture
def import_fixture(simple_einstein_ring_likelihood_2d):
    """

    :param simple_einstein_ring_likelihood_2d: fixture
    :return:
    """
    likelihood, kwargs_truths = simple_einstein_ring_likelihood_2d
    sampler = NautilusSampler(likelihood_module=likelihood)
    return sampler, likelihood


class TestNautilusSampler(object):
    """Test the fitting sequences."""

    def setup_method(self):
        pass

    def test_sampler(self, import_fixture):
        sampler, likelihood = import_fixture
        kwargs = {
            "mpi": False,
            "verbose": True,
            "f_live": 1.0,
            "n_eff": 0.0,
            "seed": 42,
        }
        points, log_w, log_l, log_z = sampler.run(**kwargs)
        assert len(points) == len(log_w)
        assert len(points) == len(log_l)
        assert np.isfinite(log_z)


if __name__ == "__main__":
    pytest.main()



__author__ = "aymgal, johannesulf"

import numpy as np

from lenstronomy.Sampling.Samplers.base_nested_sampler import NestedSampler
import lenstronomy.Util.sampling_util as utils

__all__ = ["DynestySampler"]


class DynestySampler(NestedSampler):
    """Wrapper for dynamical nested sampling algorithm Dynesty by J. Speagle.

    paper : https://arxiv.org/abs/1904.02180
    doc : https://dynesty.readthedocs.io/
    """

    def __init__(
        self,
        likelihood_module,
        prior_type="uniform",
        prior_means=None,
        prior_sigmas=None,
        width_scale=1,
        sigma_scale=1,
        bound="multi",
        sample="auto",
        use_mpi=False,
        use_pool=None,
    ):
        """
        :param likelihood_module: likelihood_module like in likelihood.py (should be callable)
        :param prior_type: 'uniform' of 'gaussian', for converting the unit hypercube to param cube
        :param prior_means: if prior_type is 'gaussian', mean for each param
        :param prior_sigmas: if prior_type is 'gaussian', std dev for each param
        :param width_scale: scale the widths of the parameters space by this factor
        :param sigma_scale: if prior_type is 'gaussian', scale the gaussian sigma by this factor
        :param bound: specific to Dynesty, see https://dynesty.readthedocs.io
        :param sample: specific to Dynesty, see https://dynesty.readthedocs.io
        :param use_mpi: Use MPI computing if `True`
        :param use_pool: specific to Dynesty, see https://dynesty.readthedocs.io
        """
        self._check_install()
        super(DynestySampler, self).__init__(
            likelihood_module,
            prior_type,
            prior_means,
            prior_sigmas,
            width_scale,
            sigma_scale,
        )

        # create the Dynesty sampler
        if use_mpi:
            from schwimmbad import MPIPool
            import sys

            # use_dill=True not supported for some versions of schwimmbad
            pool = MPIPool(use_dill=True)
            if not pool.is_master():
                pool.wait()
                sys.exit(0)

            self._sampler = self._dynesty.DynamicNestedSampler(
                loglikelihood=self.log_likelihood,
                prior_transform=self.prior,
                ndim=self.n_dims,
                bound=bound,
                sample=sample,
                pool=pool,
                use_pool=use_pool,
            )
        else:
            self._sampler = self._dynesty.DynamicNestedSampler(
                loglikelihood=self.log_likelihood,
                prior_transform=self.prior,
                ndim=self.n_dims,
                bound=bound,
                sample=sample,
            )

    def run(self, kwargs_run):
        """Run the Dynesty nested sampler.

        see https://dynesty.readthedocs.io for content of kwargs_run

        :param kwargs_run: kwargs directly passed to DynamicNestedSampler.run_nested
        :return: samples, means, logZ, logZ_err, logL, results
        """
        print("prior type :", self.prior_type)
        print("parameter names :", self.param_names)

        self._sampler.run_nested(**kwargs_run)

        results = self._sampler.results
        samples_w = results.samples  # weighted samples
        log_l = results.logl
        log_z = results.logz
        log_z_err = results.logzerr

        # Compute weighted mean and covariance.
        weights = np.exp(results.logwt - log_z[-1])  # normalized weights
        if np.sum(weights) != 1.0:
            # TODO : clearly this is not optimal...
            # weights should by definition be normalized, but it appears that for very small
            # number of live points (typically in test routines),
            # it is not *quite* the case (up to 6 decimals)
            weights = weights / np.sum(weights)

        means, covs = self._dyfunc.mean_and_cov(samples_w, weights)

        # Resample weighted samples to get equally weighted (aka unweighted) samples
        samples = self._dyfunc.resample_equal(samples_w, weights)

        return samples, means, log_z, log_z_err, log_l, results

    def _check_install(self):
        try:
            import dynesty
            import dynesty.utils as dyfunc
        except ImportError:
            print(
                "Warning : dynesty not properly installed (results might be unexpected). \
                    You can get it with $pip install dynesty."
            )
            self._dynesty_installed = False
        else:
            self._dynesty_installed = True
            self._dynesty = dynesty
            self._dyfunc = dyfunc



__author__ = ["nataliehogg", "pierrefleury", "danjohnson98"]

from lenstronomy.LensModel.single_plane import SinglePlane
import numpy as np
import copy

__all__ = ["SinglePlaneLOS"]


class SinglePlaneLOS(SinglePlane):
    """This class is based on the 'SinglePlane' class, modified to include line-of-sight
    effects as presented by Fleury et al. in 2104.08883.

    Are modified:
    - init (to include a new attribute, self.los)
    - fermat potential
    - alpha
    - hessian

    Are unchanged (inherited from SinglePlane):
    - ray_shooting, because it calls the modified alpha
    - mass_2d, mass_3d, density which refer to the main lens without LOS
    corrections.
    """

    def __init__(
        self,
        lens_model_list,
        index_los,
        numerical_alpha_class=None,
        lens_redshift_list=None,
        z_source_convention=None,
        kwargs_interp=None,
        kwargs_synthesis=None,
    ):
        """
        Instance of SinglePlaneLOS() based on the SinglePlane(), except:
        - argument "index_los" indicating the position of the LOS model in the
        lens_model_list (for correct association with kwargs)
        - attribute "los" containing the LOS model.
        :param kwargs_synthesis: keyword arguments for the 'SYNTHESIS' lens model, if applicable
        """

        super(SinglePlaneLOS, self).__init__(lens_model_list)
        # NB: It is important to run that init first, in order to create a
        # list_func for the entire model, before splitting it between a main
        # lens and the LOS corrections

        # Extract the los model and import its class
        self._index_los = index_los
        self._los_model = lens_model_list[index_los]
        self.los = self._import_class(
            self._los_model,
            custom_class=None,
            kwargs_interp=None,
            kwargs_synthesis=kwargs_synthesis,
        )

        # Define a separate class for the main lens
        lens_model_list_wo_los = [
            model for i, model in enumerate(lens_model_list) if i != index_los
        ]
        self._main_lens = SinglePlane(
            lens_model_list_wo_los,
            numerical_alpha_class=numerical_alpha_class,
            lens_redshift_list=lens_redshift_list,
            z_source_convention=z_source_convention,
            kwargs_interp=kwargs_interp,
            kwargs_synthesis=kwargs_synthesis,
        )

    def split_lens_los(self, kwargs):
        """This function splits the list of key-word arguments given to the lens model
        into those that correspond to the lens itself (kwargs_main), and those that
        correspond to the line-of-sight corrections (kwargs_los).

        :param kwargs: the list of key-word arguments passed to lenstronomy
        :return: a list of kwargs corresponding to the lens and a list of kwargs
            corresponding to the LOS effects
        """

        kwargs_los = copy.deepcopy(kwargs[self._index_los])
        # if 'LOS_MINIMAL' is at play, we set Gamma_os = Gamma_los
        # and Gamma_ds = Gamma_od
        if self._los_model == "LOS_MINIMAL":
            kwargs_los["kappa_os"] = kwargs_los.pop("kappa_los")
            kwargs_los["gamma1_os"] = kwargs_los.pop("gamma1_los")
            kwargs_los["gamma2_os"] = kwargs_los.pop("gamma2_los")
            kwargs_los["omega_os"] = kwargs_los.pop("omega_los")
            kwargs_los["kappa_ds"] = kwargs_los["kappa_od"]
            kwargs_los["gamma1_ds"] = kwargs_los["gamma1_od"]
            kwargs_los["gamma2_ds"] = kwargs_los["gamma2_od"]
            kwargs_los["omega_ds"] = kwargs_los["omega_od"]

        kwargs_main = [kwarg for i, kwarg in enumerate(kwargs) if i != self._index_los]

        return kwargs_main, kwargs_los

    def fermat_potential(
        self, x_image, y_image, kwargs_lens, x_source=None, y_source=None, k=None
    ):
        """Calculates the Fermat Potential with LOS corrections in the tidal regime.

        :param x_image: image position
        :param y_image: image position
        :param x_source: source position
        :param y_source: source position
        :param kwargs_lens: list of keyword arguments of lens model parameters matching
            the lens model classes
        :return: fermat potential in arcsec**2 as a list
        """

        kwargs_main, kwargs_los = self.split_lens_los(kwargs_lens)

        # the amplification matrices
        A_od = np.array(
            [
                [
                    1 - kwargs_los["kappa_od"] - kwargs_los["gamma1_od"],
                    -kwargs_los["gamma2_od"] + kwargs_los["omega_od"],
                ],
                [
                    -kwargs_los["gamma2_od"] - kwargs_los["omega_od"],
                    1 - kwargs_los["kappa_od"] + kwargs_los["gamma1_od"],
                ],
            ]
        )
        A_os = np.array(
            [
                [
                    1 - kwargs_los["kappa_os"] - kwargs_los["gamma1_os"],
                    -kwargs_los["gamma2_os"] + kwargs_los["omega_os"],
                ],
                [
                    -kwargs_los["gamma2_os"] - kwargs_los["omega_os"],
                    1 - kwargs_los["kappa_os"] + kwargs_los["gamma1_os"],
                ],
            ]
        )
        A_ds = np.array(
            [
                [
                    1 - kwargs_los["kappa_ds"] - kwargs_los["gamma1_ds"],
                    -kwargs_los["gamma2_ds"] + kwargs_los["omega_ds"],
                ],
                [
                    -kwargs_los["gamma2_ds"] - kwargs_los["omega_ds"],
                    1 - kwargs_los["kappa_ds"] + kwargs_los["gamma1_ds"],
                ],
            ]
        )

        # the inverse and transposed amplification matrices
        A_od_tsp = np.transpose(A_od)
        A_ds_inv = np.linalg.inv(A_ds)
        A_os_inv = np.linalg.inv(A_os)

        # the composite amplification matrices
        A_LOS = np.dot(np.dot(A_od_tsp, A_ds_inv), A_os)

        # Angular position where the ray hits the deflector's plane
        x_d, y_d = self.los.distort_vector(
            x_image,
            y_image,
            kappa=kwargs_los["kappa_od"],
            omega=kwargs_los["omega_od"],
            gamma1=kwargs_los["gamma1_od"],
            gamma2=kwargs_los["gamma2_od"],
        )

        # Evaluating the potential of the main lens at this position
        effective_potential = self._main_lens.potential(
            x_d, y_d, kwargs=kwargs_main, k=k
        )

        # obtaining the source position
        if x_source is None or y_source is None:
            x_source, y_source = self.ray_shooting(x_image, y_image, kwargs_lens, k=k)

        # the source position, modified by A_os_inv
        b_x = A_os_inv[0][0] * x_source + A_os_inv[0][1] * y_source
        b_y = A_os_inv[1][0] * x_source + A_os_inv[1][1] * y_source

        # alpha'
        f_x = x_image - b_x
        f_y = y_image - b_y

        # alpha' must then be further distorted by A_LOS
        a_x = A_LOS[0][0] * f_x + A_LOS[0][1] * f_y
        a_y = A_LOS[1][0] * f_x + A_LOS[1][1] * f_y

        # we can then obtain the geometrical term
        geometry = (f_x * a_x + f_y * a_y) / 2

        return geometry - effective_potential

    def alpha(self, x, y, kwargs, k=None):
        """Displacement angle including the line-of-sight corrections.

        :param x: x-position (preferentially arcsec)
        :type x: numpy array
        :param y: y-position (preferentially arcsec)
        :type y: numpy array
        :param kwargs: list of keyword arguments of lens model parameters matching the
            lens model classes, including line-of-sight corrections
        :param k: only evaluate the k-th lens model
        :return: deflection angles in units of arcsec
        """

        kwargs_main, kwargs_los = self.split_lens_los(kwargs)

        # Angular position where the ray hits the deflector's plane
        x_d, y_d = self.los.distort_vector(
            x,
            y,
            kappa=kwargs_los["kappa_od"],
            omega=kwargs_los["omega_od"],
            gamma1=kwargs_los["gamma1_od"],
            gamma2=kwargs_los["gamma2_od"],
        )

        # Displacement due to the main lens only
        f_x, f_y = self._main_lens.alpha(x_d, y_d, kwargs=kwargs_main, k=k)

        # Correction due to the background convergence, shear and rotation
        f_x, f_y = self.los.distort_vector(
            f_x,
            f_y,
            kappa=kwargs_los["kappa_ds"],
            omega=kwargs_los["omega_ds"],
            gamma1=kwargs_los["gamma1_ds"],
            gamma2=kwargs_los["gamma2_ds"],
        )

        # Perturbed position in the absence of the main lens
        x_os, y_os = self.los.distort_vector(
            x,
            y,
            kappa=kwargs_los["kappa_os"],
            omega=kwargs_los["omega_os"],
            gamma1=kwargs_los["gamma1_os"],
            gamma2=kwargs_los["gamma2_os"],
        )

        # Complete displacement
        f_x += x - x_os
        f_y += y - y_os

        return f_x, f_y

    def hessian(self, x, y, kwargs, k=None):
        """Hessian matrix.

        :param x: x-position (preferentially arcsec)
        :type x: numpy array
        :param y: y-position (preferentially arcsec)
        :type y: numpy array
        :param kwargs: list of keyword arguments of lens model parameters matching the
            lens model classes
        :param k: only evaluate the k-th lens model
        :return: f_xx, f_xy, f_yx, f_yy components
        """

        kwargs_main, kwargs_los = self.split_lens_los(kwargs)

        # Angular position where the ray hits the deflector's plane
        x_d, y_d = self.los.distort_vector(
            x,
            y,
            kappa=kwargs_los["kappa_od"],
            omega=kwargs_los["omega_od"],
            gamma1=kwargs_los["gamma1_od"],
            gamma2=kwargs_los["gamma2_od"],
        )

        # Hessian matrix of the main lens only
        f_xx, f_xy, f_yx, f_yy = self._main_lens.hessian(
            x_d, y_d, kwargs=kwargs_main, k=k
        )

        # Multiply on the left by (1 - Gamma_ds)
        f_xx, f_xy, f_yx, f_yy = self.los.left_multiply(
            f_xx,
            f_xy,
            f_yx,
            f_yy,
            kappa=kwargs_los["kappa_ds"],
            omega=kwargs_los["omega_ds"],
            gamma1=kwargs_los["gamma1_ds"],
            gamma2=kwargs_los["gamma2_ds"],
        )

        # Multiply on the right by (1 - Gamma_od)
        f_xx, f_xy, f_yx, f_yy = self.los.right_multiply(
            f_xx,
            f_xy,
            f_yx,
            f_yy,
            kappa=kwargs_los["kappa_od"],
            omega=kwargs_los["omega_od"],
            gamma1=kwargs_los["gamma1_od"],
            gamma2=kwargs_los["gamma2_od"],
        )

        # LOS contribution in the absence of the main lens
        f_xx += kwargs_los["kappa_os"] + kwargs_los["gamma1_os"]
        f_xy += kwargs_los["gamma2_os"] - kwargs_los["omega_os"]
        f_yx += kwargs_los["gamma2_os"] + kwargs_los["omega_os"]
        f_yy += kwargs_los["kappa_os"] - kwargs_los["gamma1_os"]

        return f_xx, f_xy, f_yx, f_yy

    def mass_3d(self, r, kwargs, bool_list=None):
        """Computes the mass within a 3d sphere of radius r *for the main lens only*

        :param r: radius (in angular units)
        :param kwargs: list of keyword arguments of lens model parameters matching the
            lens model classes
        :param bool_list: list of bools that are part of the output
        :return: mass (in angular units, modulo epsilon_crit)
        """

        print("Note: The computation of the 3d mass ignores the LOS corrections.")

        kwargs_main, kwargs_los = self.split_lens_los(kwargs)
        mass_3d = self._main_lens.mass_3d(r=r, kwargs=kwargs_main, bool_list=bool_list)

        return mass_3d

    def mass_2d(self, r, kwargs, bool_list=None):
        """Computes the mass enclosed a projected (2d) radius r *for the main lens only*

        The mass definition is such that:

        .. math::
            \\alpha = mass_2d / r / \\pi

        with alpha is the deflection angle

        :param r: radius (in angular units)
        :param kwargs: list of keyword arguments of lens model parameters matching the lens model classes
        :param bool_list: list of bools that are part of the output
        :return: projected mass (in angular units, modulo epsilon_crit)
        """

        print("Note: The computation of the 2d mass ignores the LOS corrections.")

        kwargs_main, kwargs_los = self.split_lens_los(kwargs)
        mass_2d = self._main_lens.mass_2d(r=r, kwargs=kwargs_main, bool_list=bool_list)

        return mass_2d

    def density(self, r, kwargs, bool_list=None):
        """3d mass density at radius r *for the main lens only* The integral in the LOS
        projection of this quantity results in the convergence quantity.

        :param r: radius (in angular units)
        :param kwargs: list of keyword arguments of lens model parameters matching the
            lens model classes
        :param bool_list: list of bools that are part of the output
        :return: mass density at radius r (in angular units, modulo epsilon_crit)
        """

        print("Note: The computation of the density ignores the LOS corrections.")

        kwargs_main, kwargs_los = self.split_lens_los(kwargs)
        density = self._main_lens.density(r=r, kwargs=kwargs_main, bool_list=bool_list)

        return density

    def potential(self, x, y, kwargs, k=None):
        """Lensing potential *of the main lens only* In the presence of LOS corrections,
        the system generally does not admit a potential, in the sense that the curl of
        alpha is generally non-zero.

        :param x: x-position (preferentially arcsec)
        :type x: numpy array
        :param y: y-position (preferentially arcsec)
        :type y: numpy array
        :param kwargs: list of keyword arguments of lens model parameters matching the
            lens model classes
        :param k: only evaluate the k-th lens model
        :return: lensing potential in units of arcsec^2
        """

        print(
            "Note: The computation of the potential ignores the LOS corrections.\
              In the presence of LOS corrections, a lensing system does not always\
              derive from a potential."
        )

        # kwargs_main, kwargs_los = self.split_lens_los(kwargs)
        potential = self._main_lens.potential(x, y, kwargs, k=k)

        return potential
